import argparse, os, sys
from functools import partial
from pathlib import Path
from typing import Optional
from multiprocessing import Pool
import numpy as np
import pandas as pd
import re

from cwas.runnable import Runnable

from cwas.burden_test import BurdenTest
from cwas.utils.log import print_progress, print_arg
from cwas.utils.check import check_num_proc
from cwas.core.burden_test.binomial import binom_two_tail

class PermutationTest(BurdenTest):
    def __init__(self, args: Optional[argparse.Namespace] = None):
        super().__init__(args)
        self._perm_rrs = None
        self._binom_pvals = None
        self._perm_rrs_path = None
        self._binom_pvals_path = None

    @staticmethod
    def _print_args(args: argparse.Namespace):
        super(PermutationTest, PermutationTest)._print_args(args)
        print_arg(f"Number of permutations", args.num_perm)
        print_arg(f"Number of processes", args.num_proc)
        print_arg(f"Generate binomial p values for burden-shifted data", args.burden_shift)
        #print_arg(f"Generate relative risks (RRs) for burden-shifted data", args.save_perm_rr)

    @staticmethod
    def _check_args_validity(args: argparse.Namespace):
        super(PermutationTest, PermutationTest)._check_args_validity(args)
        check_num_proc(args.num_proc)
    
    @property
    def cat_path(self) -> Path:
        return self.args.cat_path.resolve()

    @property
    def output_dir_path(self) -> Path:
        return self.args.output_dir_path.resolve()

    @property
    def result_path(self) -> Path:
        f_name = re.sub(r'categorization_result\.txt\.gz|categorization_result\.txt', 'permutation_test.txt.gz', self.cat_path.name)
        self._result_path = Path(
            f"{self.output_dir_path}/"
            f"{f_name}"
        )
        return self._result_path
    
    @property
    def perm_rrs_path(self) -> Path:
        if self._perm_rrs_path is None:
            f_name = re.sub(r'categorization_result\.txt\.gz|categorization_result\.txt', 'permutation_RRs.txt.gz', self.cat_path.name)
            self._perm_rrs_path = Path(
                f"{self.output_dir_path}/"
                f"{f_name}"
            )
        return self._perm_rrs_path
    
    @property
    def binom_pvals_path(self) -> Path:
        if self._binom_pvals_path is None:
            f_name = re.sub(r'categorization_result\.txt\.gz|categorization_result\.txt', 'binom_pvals.txt.gz', self.cat_path.name)
            self._binom_pvals_path = Path(
                f"{self.output_dir_path}/"
                f"{f_name}"
            )
        return self._binom_pvals_path
    
    @property
    def burden_shift(self) -> bool:
        return self.args.burden_shift

    #@property
    #def save_perm_rr(self) -> bool:
    #    return self.args.save_perm_rr

    @property
    def use_n_carrier(self) -> bool:
        return self.args.use_n_carrier

    def run_burden_test(self):
        print_progress("Run permutation test")

        vals = np.concatenate(
            self.cal_perm_rr(self.categorization_result,
                             self.args.num_perm,
                             burden_shift=self.burden_shift)
        )
        
        if self.burden_shift:
            perm_rrs = vals[range(0, len(vals), 2)]
            binom_pvals = vals[range(1, len(vals), 2)]
        else: 
            perm_rrs = vals
            
        self._result["P"] = self.get_perm_pval(
            perm_rrs,
            rr = self._result["Relative_Risk"].values
        )

        low_P_idx = self._result[self._result["P"] < 0.01].index
        print_progress(f"Run additional permutation test for {len(low_P_idx)} categories with P < 0.01")
        perm_rrs_x10 = np.concatenate(
            self.cal_perm_rr(
                self.categorization_result[low_P_idx],
                10*self.args.num_perm,
                burden_shift=False,
            )
        )

        self._result.loc[low_P_idx, "P"] = self.get_perm_pval(
            perm_rrs_x10,
            rr = self._result.loc[low_P_idx]["Relative_Risk"].values
        )
        ## Make a dataframe of permutation RRs
        #if self.save_perm_rr:
        #    self._perm_rrs = pd.DataFrame(perm_rrs, columns=self.categorization_result.columns)
        #    self._perm_rrs.index += 1
        #    self._perm_rrs.index.name = 'Trial'
        
        ## Make a dataframe of binomial p values
        if self.burden_shift:
            self._binom_pvals = pd.DataFrame(binom_pvals, columns=self.categorization_result.columns)
            self._binom_pvals.index += 1
            self._binom_pvals.index.name = 'Trial'

    def cal_perm_rr(self, categorization_result: pd.DataFrame, num_perm: int, burden_shift: bool) -> np.ndarray:
        print_progress(f"Calculate permutation RRs (# of permutations: {num_perm})")

        var_counts = categorization_result[np.isin(self.phenotypes, ['case', 'ctrl'])].values
        
        _burden_test_partial = partial(self._burden_test, 
                                       case_cnt=self.case_cnt,
                                       ctrl_cnt=self.ctrl_cnt,
                                       var_counts=var_counts,
                                       use_n_carrier=self.use_n_carrier,
                                       burden_shift=burden_shift)

        if self.args.num_proc == 1:
            array_list = _burden_test_partial(
                (0, num_perm),
            )
        else:
            seed_range = []
            range_len = num_perm // self.args.num_proc
            if range_len == 0:
                raise AssertionError(f'The number of processors ("{self.args.num_proc:,d}") are larger than '
                                        f'the number of permutations ("{self.args.num_proc:,d}").')

            for i in range(self.args.num_proc - 1):
                r = (range_len * i, range_len * (i + 1))
                seed_range.append(r)
            seed_range.append((range_len * (self.args.num_proc - 1), num_perm))
            def mute():
                sys.stderr = open(os.devnull, 'w')  
            with Pool(self.args.num_proc, initializer=mute) as pool:
                sub_lists = pool.map(_burden_test_partial, seed_range)
            array_list = []
            for sub_list in sub_lists:
                array_list.extend(sub_list)

        return array_list
    
    @staticmethod
    def _burden_test(seed_range: tuple, case_cnt: int, ctrl_cnt: int, var_counts: np.ndarray, use_n_carrier: bool, burden_shift: bool):
        array_list = []
        total_cnt = case_cnt + ctrl_cnt
        for seed in range(10001 + seed_range[0], 10001 + seed_range[1]):
            ## Make an array for random swapping
            swap_labels = np.full(total_cnt, 'ctrl')
            ## For reproducibility
            np.random.seed(seed=seed)
            ## Make an index for random swapping (variables in the index will be cases)
            idx = np.random.choice(range(0,total_cnt), case_cnt, replace=False)
            
            for k in idx:
                swap_labels[k] = 'case'

            are_case = swap_labels == 'case'
            
            if use_n_carrier:
                is_carrier = np.where(var_counts > 0, 1, 0)
                n1 = is_carrier[are_case, :].sum(axis=0)
                n2 = is_carrier[~are_case, :].sum(axis=0)
            else:
                n1 = var_counts[are_case, :].sum(axis=0).round()
                n2 = var_counts[~are_case, :].sum(axis=0).round()
                
            norm_n1 = n1 / case_cnt
            norm_n2 = n2 / ctrl_cnt
            perm_rr = norm_n1 / norm_n2
            
            binom_p = case_cnt / total_cnt
            
            ## Calculate binomial p values for burden-shifted data
            if burden_shift:
                binom_pval = np.vectorize(binom_two_tail)(
                    n1, n2, binom_p
                )
                ## If RR is below 1, the sign of binomial p value is reversed
                binom_pval[perm_rr < 1] *= -1
                array_list.append(np.concatenate((perm_rr[np.newaxis, :], binom_pval[np.newaxis, :])))
            else:
                array_list.append(perm_rr[np.newaxis, :])
        return array_list

    def get_perm_pval(self, perm_rrs, rr: np.ndarray):
        ## Permutation tests
        ## Check whether a permutation RR is more extreme than the original RR
        are_ext_rr = ((rr >= 1) & (perm_rrs >= rr)) | ((rr < 1) & (perm_rrs <= rr))
        ext_rr_cnt = are_ext_rr.sum(axis=0)
        return (ext_rr_cnt + 1) / (are_ext_rr.shape[0] + 1)
        
    def save_result(self):
        super().save_result()
        #if self.save_perm_rr:
        #    print_progress(f"Save the permutation RRs to the file {self.perm_rrs_path}")
        #    self._perm_rrs.to_csv(self.perm_rrs_path, sep='\t', compression='gzip')
        if self.burden_shift:
            print_progress(f"Save the binomial p values to the file {self.binom_pvals_path}")
            self._binom_pvals.to_csv(self.binom_pvals_path, sep='\t', compression='gzip')


from scipy.stats import binom_test


def binom_two_tail(n1: int, n2: int, p: float):
    return binom_test(x=n1, n=n1 + n2, p=p, alternative="two-sided")


def binom_one_tail(n1: int, n2: int, p: float):
    return binom_test(x=n1, n=n1 + n2, p=p, alternative="greater")



"""
Functions to check common conditions that cause cumbersome errors in this
project. Those conditions should be checked to prevent this project from
being shut down unexpectedly.
"""
import multiprocessing as mp
from pathlib import Path
from typing import Union


def check_is_file(file_path: Union[Path, str]):
    if not file_path:
        raise ValueError("The file path cannot be None.")

    if not isinstance(file_path, Path):
        file_path = Path(file_path)

    if not file_path.is_file():
        raise FileNotFoundError(
            f'"{file_path}" cannot be found. Check this file path first.'
        )


def check_is_dir(dir_path: Union[Path, str]):
    if not dir_path:
        raise ValueError(f"dir_path: {dir_path}")

    if not isinstance(dir_path, Path):
        dir_path = Path(dir_path)

    if not dir_path.is_dir():
        raise NotADirectoryError(
            f'"{dir_path}" is not a valid directory. Check this path first.'
        )


def check_num_proc(num_proc: int):
    max_num_proc = mp.cpu_count()

    if num_proc < 1:
        raise ValueError(
            f"Your number of worker processes '{num_proc}' is invalid. "
            f"This number must be a positive integer."
        )
    elif num_proc > max_num_proc:
        raise ValueError(
            f"Your number of worker processes '{num_proc}' exceeds the "
            f"number of available CPUs in your computer ({max_num_proc})."
        )

def check_same_n_lines(file_paths: list, gzip_file: bool = False):
    n_lines_set = set()
    for file_path in file_paths:
        f = gzip.open(file_path, "r") if gzip_file else open(file_path, "r")
        for count, _ in enumerate(f):
                pass
        f.close()
        n_lines_set.add(count)
    if len(n_lines_set) > 1:
        raise ValueError(
            f"Some files in the previous step have different number of lines. "
            f"Check the files first."
        )



"""
Test cwas.core.configuration.create
"""
import cwas.core.configuration.create as create
import pytest
import yaml


@pytest.fixture(scope="module", autouse=True)
def setup(cwas_workspace):
    cwas_workspace.mkdir()


@pytest.fixture(scope="module", autouse=True)
def teardown(cwas_workspace):
    yield
    remove_workspace(cwas_workspace)


def remove_workspace(cwas_workspace):
    for f in cwas_workspace.glob("*"):
        f.unlink()
    cwas_workspace.rmdir()


def test_create_annotation_key_bed(cwas_workspace, annotation_dir):
    bed_key_conf = cwas_workspace / "annotation_keys.yaml"
    create.create_annotation_key(bed_key_conf, annotation_dir, "bed")

    assert bed_key_conf.exists()
    with bed_key_conf.open() as f:
        bed_key = yaml.safe_load(f)
    assert "bed_annot1.bed.gz" in bed_key
    assert "bed.annot2.bed.gz" in bed_key
    assert "bed.annot.bed" not in bed_key
    assert bed_key["bed_annot1.bed.gz"] == "bed_annot1"
    assert bed_key["bed.annot2.bed.gz"] == "bed_annot2"

    bed_key_conf.unlink()


def test_create_category_domain_list(
    cwas_workspace, annotation_key_conf, gene_matrix
):
    # Setting
    bed_key_conf = cwas_workspace / "annotation_keys.yaml"
    create.split_annotation_key(bed_key_conf, annotation_key_conf)

    domain_list_path = cwas_workspace / "category_domain.yaml"
    create.create_category_domain_list(
        domain_list_path, bed_key_conf, gene_matrix
    )

    assert domain_list_path.is_file()

    with domain_list_path.open("r") as domain_list_f, bed_key_conf.open(
        "r"
    ) as bed_key_f, gene_matrix.open(
        "r"
    ) as gene_mat_f:
        domain_dict = yaml.safe_load(domain_list_f)
        bed_key_dict = yaml.safe_load(bed_key_f)
        region_dict = bed_key_dict['functional_annotation']
        score_dict = bed_key_dict['functional_score']
        gene_mat_header = gene_mat_f.readline()
        gene_list_domains = gene_mat_header.strip().split()
        gene_list_domains = gene_list_domains[2:]

    assert len(domain_dict["region"]) == 1 + len(region_dict)
    assert len(domain_dict["conservation"]) == 1 + len(score_dict)
    assert len(domain_dict["gene_list"]) == 1 + len(gene_list_domains)

    bed_key_conf.unlink()
    domain_list_path.unlink()


def test_create_redundant_category_table(cwas_workspace):
    redundant_category_table = cwas_workspace / "redundant_category.txt"
    create.create_redundant_category_table(redundant_category_table)
    assert redundant_category_table.exists()
    redundant_category_table.unlink()


def test__load_yaml_file_error(cwas_workspace):
    test_path = cwas_workspace / "not_yaml.txt"
    with test_path.open("w") as test_f:
        print('{Hello: World!}"', file=test_f)
    with pytest.raises(yaml.YAMLError):
        _ = create.split_annotation_key(None, None, test_path)
    test_path.unlink()



import pandas as pd
import pytest
from cwas.binomial_test import BinomialTest


class BinomialTestMock(BinomialTest):
    """This class do not make outputs"""

    def save_result(self):
        pass

    def update_env(self):
        pass


@pytest.fixture
def categorization_result():
    results = [
        {"SAMPLE": "Sample1", "A_B_C_D_E": 5, "a_b_c_d_e": 4},
        {"SAMPLE": "Sample2", "A_B_C_D_E": 10, "a_b_c_d_e": 16},
        {"SAMPLE": "Sample3", "A_B_C_D_E": 12, "a_b_c_d_e": 8},
        {"SAMPLE": "Sample4", "A_B_C_D_E": 7, "a_b_c_d_e": 11},
        {"SAMPLE": "Sample5", "A_B_C_D_E": 8, "a_b_c_d_e": 6},
        {"SAMPLE": "Sample6", "A_B_C_D_E": 15, "a_b_c_d_e": 10},
    ]
    return pd.DataFrame(results).set_index("SAMPLE")


@pytest.fixture
def sample_info():
    samples = [
        {"SAMPLE": "Sample1", "FAMILY": "F1", "PHENOTYPE": "case"},
        {"SAMPLE": "Sample2", "FAMILY": "F1", "PHENOTYPE": "ctrl"},
        {"SAMPLE": "Sample3", "FAMILY": "F2", "PHENOTYPE": "case"},
        {"SAMPLE": "Sample4", "FAMILY": "F2", "PHENOTYPE": "ctrl"},
        {"SAMPLE": "Sample5", "FAMILY": "F3", "PHENOTYPE": "ctrl"},
        {"SAMPLE": "Sample6", "FAMILY": "F3", "PHENOTYPE": "ctrl"},
    ]
    return pd.DataFrame(samples).set_index("SAMPLE")


@pytest.fixture
def adjustment_factor():
    adj_factors = [
        {"SAMPLE": "Sample1", "AdjustFactor": 2.0},
        {"SAMPLE": "Sample2", "AdjustFactor": 0.5},
        {"SAMPLE": "Sample3", "AdjustFactor": 0.25},
        {"SAMPLE": "Sample4", "AdjustFactor": 1.0},
        {"SAMPLE": "Sample5", "AdjustFactor": 0.5},
        {"SAMPLE": "Sample6", "AdjustFactor": 0.2},
    ]
    return pd.DataFrame(adj_factors).set_index("SAMPLE")


@pytest.fixture
def sample_info_other_sample():
    samples = [
        {"SAMPLE": "SampleA", "FAMILY": "F1", "PHENOTYPE": "case"},
        {"SAMPLE": "SampleB", "FAMILY": "F2", "PHENOTYPE": "ctrl"},
    ]
    return pd.DataFrame(samples).set_index("SAMPLE")


@pytest.fixture
def adjustment_factor_other_sample():
    adj_factors = [
        {"SAMPLE": "SampleA", "AdjustFactor": 2.0},
        {"SAMPLE": "SampleB", "AdjustFactor": 0.5},
    ]
    return pd.DataFrame(adj_factors).set_index("SAMPLE")


@pytest.fixture
def binomial_test(
    categorization_result, sample_info, adjustment_factor,
):
    # This is not an appropriate usage.
    inst = BinomialTestMock()
    inst._categorization_result = categorization_result
    inst._sample_info = sample_info
    inst._adj_factor = adjustment_factor
    return inst


@pytest.fixture
def binomial_test_with_inconsistent_sample(
    categorization_result,
    sample_info_other_sample,
    adjustment_factor_other_sample,
):
    # This is not an appropriate usage.
    inst = BinomialTestMock()
    inst._categorization_result = categorization_result
    inst._sample_info = sample_info_other_sample
    inst._adj_factor = adjustment_factor_other_sample
    return inst


def test_adjust_categorization_result(binomial_test):
    binomial_test._adjust_categorization_result()
    categorization_result = binomial_test.categorization_result
    assert categorization_result.loc["Sample1"].to_list() == [10, 8]
    assert categorization_result.loc["Sample2"].to_list() == [5, 8]
    assert categorization_result.loc["Sample3"].to_list() == [3, 2]
    assert categorization_result.loc["Sample4"].to_list() == [7, 11]
    assert categorization_result.loc["Sample5"].to_list() == [4, 3]
    assert categorization_result.loc["Sample6"].to_list() == [3, 2]


def test_adjust_categorization_with_inconsistent_sample(
    binomial_test_with_inconsistent_sample,
):
    with pytest.raises(ValueError):
        binomial_test_with_inconsistent_sample._adjust_categorization_result()


def test_run_with_inconsistent_sample(binomial_test_with_inconsistent_sample):
    with pytest.raises(ValueError):
        binomial_test_with_inconsistent_sample.run()


def test_run(binomial_test):
    binomial_test._adjust_categorization_result()
    binomial_test.run()
    assert binomial_test._result is not None
    assert binomial_test._result.index.name == "Category"
    expected_columns = [
        "variant_type",
        "gene_list",
        "conservation",
        "gencode",
        "region",
        "Case_DNV_Count",
        "Ctrl_DNV_Count",
        "Relative_Risk",
        "P",
        "P_1side",
        "Z_1side",
    ]
    expected_index = [
        "A_B_C_D_E",
        "a_b_c_d_e",
    ]
    assert list(binomial_test._result.columns.values) == expected_columns
    assert list(binomial_test._result.index.values) == expected_index


def test_binom_p(binomial_test):
    assert binomial_test.binom_p == 1 / 3  # A fraction of cases


def test_case_cnt(binomial_test):
    assert binomial_test.case_cnt == 2


def test_ctrl_cnt(binomial_test):
    assert binomial_test.ctrl_cnt == 4


def test_case_variant_cnt(binomial_test):
    binomial_test._adjust_categorization_result()
    assert list(binomial_test.case_variant_cnt) == [13, 10]


def test_ctrl_variant_cnt(binomial_test):
    binomial_test._adjust_categorization_result()
    assert list(binomial_test.ctrl_variant_cnt) == [19, 24]


def test_calculate_relative_risk(binomial_test):
    binomial_test._adjust_categorization_result()
    binomial_test.run()
    expected_relative_risk1 = (13 / 2) / (19 / 4)  # A_B_C_D_E
    expected_relative_risk2 = (10 / 2) / (24 / 4)  # a_b_c_d_e
    assert binomial_test._result["Relative_Risk"].to_list() == [
        expected_relative_risk1,
        expected_relative_risk2,
    ]




"""
Functions to prepare CWAS annotation step
"""
from __future__ import annotations

import multiprocessing as mp
from collections import defaultdict
from functools import partial
from pathlib import Path
from typing import Generator

import cwas.utils.log as log
import numpy as np
from cwas.core.preparation.bedreader import BedReader


def merge_bed_files(
    out_merge_bed: Path,
    bed_file_and_keys: list[tuple[Path, str]],
    num_proc: int = 1,
    force_overwrite: int = 0,
):
    """ Merge all information of coordinates from all the BED files
    into one file.

    For example, assume that following three coordinates exist
    in bed file A, B and C, respectively.
    - chr1  100 300 (from A BED file)
    - chr1  200 400 (from B BED file)
    - chr1  250 350 (from C BED file)

    If these coordinates are merged,
    than following coordinates will be newly produced.
    - chr1  100 200 annotation A
    - chr1  200 250 annotation A, B
    - chr1  250 300 annotation A, B, C
    - chr1  300 350 annotation B, C
    - chr1  350 400 annotation B

    The annotation information for each coordinate will be represented
    as an annotation integer, which is a one-hot encoding of annotation
    information. In this case, A and C are represented as a least significant
    bit and a most significant bit, respectively.

    For example, if all the BED files are A, B, and C,
    then an annotation integer of the second coordinate above is 3(b'011).

    So here is a final result.
    - chr1  100 200 1 (b'001)
    - chr1  200 250 3 (b'011)
    - chr1  250 300 7 (b'111)
    - chr1  300 350 6 (b'110)
    - chr1  350 400 2 (b'010)

    """
    bed_gz_path = Path(str(out_merge_bed) + ".gz")
    if not force_overwrite and (out_merge_bed.exists() or bed_gz_path.exists()):
        log.print_warn("Merged BED file already exists. Skip merging.")
        return

    tmp_dir = out_merge_bed.parent / "tmp"
    tmp_dir.mkdir(exist_ok=True)

    # Make temporary merged BED paths for each chromosome
    chroms = [f"chr{n}" for n in range(1, 23)]
    merge_bed_paths = {}

    for chrom in chroms:
        merge_bed_name = out_merge_bed.name.replace(".bed", f".{chrom}.bed")
        merge_bed_paths[chrom] = tmp_dir / merge_bed_name

    # Make a merged BED file for each chromosome
    bed_paths, bed_ids = zip(*bed_file_and_keys)
    try:
        if num_proc == 1:
            for chrom in chroms:
                merge_bed_files_by_chrom(
                    merge_bed_paths[chrom], chrom, bed_paths, force_overwrite
                )
        else:
            pool = mp.Pool(num_proc)
            pool.starmap(
                partial(
                    merge_bed_files_by_chrom,
                    bed_file_paths=bed_paths,
                    force_overwrite=force_overwrite,
                ),
                [(merge_bed_paths[chrom], chrom) for chrom in chroms],
            )
            pool.close()
            pool.join()
    except Exception:
        log.print_err(
            "Merging BED files has failed because some error occurred."
        )
        # Remove the temporary directory if empty
        if not list(tmp_dir.glob("*")):
            log.print_progress(f'Remove "{tmp_dir}"')
            tmp_dir.rmdir()
        raise

    with out_merge_bed.open("w") as outfile:
        annot_key_str = "|".join(bed_ids)
        print(f"#ANNOT={annot_key_str}", file=outfile)
        print("#chrom", "start", "end", "annot_int", sep="\t", file=outfile)

        for chrom in chroms:
            merge_bed_path = merge_bed_paths[chrom]
            with merge_bed_path.open() as merge_bed_file:
                for line in merge_bed_file:
                    print(line, end="", file=outfile)
            merge_bed_path.unlink()

    tmp_dir.rmdir()


def merge_bed_files_by_chrom(
    out_merge_bed: Path,
    chrom: str,
    bed_file_paths: list[Path],
    force_overwrite: int = 0,
):
    if not force_overwrite and out_merge_bed.exists():
        log.print_warn(
            f'"{out_merge_bed}" already exists. '
            f"Skip merging BED files for {chrom}."
        )
        return
    try:
        log.print_progress(f"Merge BED files for {chrom}")
        _merge_bed_files(out_merge_bed, chrom, bed_file_paths)
    except Exception:
        log.print_err(
            f"Some error occurred during merging BED files for {chrom}."
        )
        if out_merge_bed.exists():
            # Clean the incomplete output
            log.print_progress(f'Remove "{out_merge_bed}"')
            out_merge_bed.unlink()
        raise


def bed_reader_iter(
    bed_file_paths: list[Path], chrom: str
) -> Generator(BedReader):
    """Generate the BedReader instance for each BED file paths"""
    for bed_file_path in bed_file_paths:
        bedreader_inst = BedReader(bed_file_path)
        bedreader_inst.set_contig(chrom)
        yield bedreader_inst


def _merge_bed_files(
    out_merge_bed: Path, chrom: str, bed_file_paths: list[Path],
):
    """ Merge annotation information of all BED coordinates of one chromosome
    from all the BED files """
    start_to_key_idx = defaultdict(list)
    end_to_key_idx = defaultdict(list)
    pos_set = set()
    bed_readers = bed_reader_iter(bed_file_paths, chrom)

    # Read coordinates from each annotation bed file
    for i, bed_reader in enumerate(bed_readers):
        for _, start, end, *_ in bed_reader:
            start_to_key_idx[start].append(i)
            end_to_key_idx[end].append(i)
            pos_set.add(start)
            pos_set.add(end)

    # Create a BED file listing new coordinates
    # with merged annotation information
    pos_list = sorted(pos_set)
    annot_cnt = np.zeros(len(bed_file_paths))
    prev_pos = -1
    n_bed = 0

    with out_merge_bed.open("w") as outfile:
        for pos in pos_list:
            if n_bed > 0:  # Make a new coordinate
                one_hot = np.vectorize(lambda x: 1 if x else 0)(annot_cnt)
                annot_int = _one_hot_to_int(one_hot)
                bed_entry = (chrom, prev_pos, pos, annot_int)
                print(*bed_entry, sep="\t", file=outfile)

            end_key_ind = end_to_key_idx.get(pos)
            start_key_ind = start_to_key_idx.get(pos)

            # This position is an end of at least one bed coordinate.
            if end_key_ind is not None:
                for end_key_idx in end_key_ind:
                    annot_cnt[end_key_idx] -= 1
                n_bed -= len(end_key_ind)

            # This position is a start of at least one bed coordinate.
            if start_key_ind is not None:
                for start_key_idx in start_key_ind:
                    annot_cnt[start_key_idx] += 1
                n_bed += len(start_key_ind)

            prev_pos = pos


def _one_hot_to_int(one_hot: np.ndarray) -> int:
    """
    Note: Index 0 is the least significant bit.
    e.g.
    [1, 1, 0, 1] -> 0'b1011 = 11
    """
    n = 0

    for i in range(len(one_hot)):
        if one_hot[i]:
            n += 2 ** i

    return n



"""
CWAS test fixtures
"""
from pathlib import Path
from random import randint

import pytest
import yaml
from cwas.env import Env

_rand_n = randint(1, 1000000)


@pytest.fixture(scope="package")
def cwas_env_path():
    _tmp_env = Path.home() / f".cwas_env_{_rand_n}"
    return _tmp_env


@pytest.fixture(scope="package", autouse=True)
def cwas_env_inst(cwas_env_path):
    env_inst = Env(cwas_env_path)
    yield
    env_inst.remove_file()


@pytest.fixture(scope="package")
def cwas_workspace():
    _tmp_workspace = Path.home() / f".cwas-test-{_rand_n}"
    return _tmp_workspace


@pytest.fixture(scope="package")
def cwas_input_dir():
    _tmp_input_dir = Path.home() / f".cwas-test-input-{_rand_n}"
    return _tmp_input_dir


@pytest.fixture(scope="package")
def annotation_dir(cwas_input_dir):
    _annotation_dir = cwas_input_dir / "annotation-data"
    return _annotation_dir


@pytest.fixture(scope="package")
def annotation_key_conf(cwas_input_dir):
    _annotation_key_conf = cwas_input_dir / "annotation_key.yaml"
    return _annotation_key_conf


@pytest.fixture(scope="package")
def bw_cutoff_conf(cwas_input_dir):
    _bw_cutoff_conf = cwas_input_dir / "user_def_bigwig_cutoff.yaml"
    return _bw_cutoff_conf


@pytest.fixture(scope="package")
def gene_matrix(cwas_input_dir):
    _gene_matrix = cwas_input_dir / "gene_matrix.txt"
    return _gene_matrix


@pytest.fixture(scope="package", autouse=True)
def create_cwas_input_dir(
    cwas_input_dir, gene_matrix, annotation_key_conf, bw_cutoff_conf
):
    cwas_input_dir.mkdir()
    create_gene_matrix(gene_matrix)
    create_annotation_key_conf(annotation_key_conf)
    create_bw_cutoff_conf(bw_cutoff_conf)
    print("[TEST] Temporary CWAS input directory has created.")
    yield
    for f in cwas_input_dir.glob("*"):
        f.unlink()
    cwas_input_dir.rmdir()
    print("[TEST] Temporary CWAS input directory has deleted.")


@pytest.fixture(scope="package", autouse=True)
def create_annotation_dir(create_cwas_input_dir, annotation_dir):
    annotation_dir.mkdir()
    annot_filepaths = [
        annotation_dir / "bed_annot1.bed.gz",
        annotation_dir / "bed.annot2.bed.gz",
        annotation_dir / "bed.annot3.bed",
        annotation_dir / "bw_annot1.bw",
        annotation_dir / "bw.annot2.bw",
        annotation_dir / "bw.annot3.bw.gz",
    ]
    for annot_filepath in annot_filepaths:
        annot_filepath.touch()
    yield
    for f in annotation_dir.glob("*"):
        f.unlink()
    annotation_dir.rmdir()


def create_gene_matrix(gene_matrix):
    gene_matrix_header = [
        "gene_id",
        "gene_name",
        "gene1",
        "gene2",
        "gene3",
        "gene4",
        "gene5",
    ]
    with gene_matrix.open("w") as out_f:
        print(*gene_matrix_header, sep="\t", file=out_f)


def create_annotation_key_conf(annotation_key_conf):
    annot_key_dict = {
        "bed_annot1.bed.gz": "bed1",
        "bed.annot2.bed.gz": "bed2",
        "bed.annot3.bed": "bed3",
        "bw_annot1.bw": "bw1",
        "bw.annot2.bw": "bw2",
        "bw.annot3.bw.gz": "bw3",
    }
    with annotation_key_conf.open("w") as out_f:
        yaml.safe_dump(annot_key_dict, out_f)


def create_bw_cutoff_conf(bw_cutoff_conf):
    bw_cutoff_dict = {
        "bw_annot1.bw": 1.0,
        "bw.annot2.bw": -2.2,
        "bw.annot3.bw.gz": 3.3,
    }
    with bw_cutoff_conf.open("w") as out_f:
        yaml.safe_dump(bw_cutoff_dict, out_f)



"""
Tests cwas.core.prepartion.annotation
"""
from __future__ import annotations

import random
from pathlib import Path

import cwas.core.preparation.annotation as annotation
import pytest
from cwas.utils.cmd import compress_using_bgzip, index_using_tabix

NUM_CHROM = random.randint(1, 22)


@pytest.fixture(scope="module")
def input_coordinates():
    """Make input coordinates without a contig (chromosome)
    """
    input_coordinates = []

    # Number of appends = Number of BED files
    input_coordinates.append(
        [(100, 300, 1), (1000, 1100, 1),]
    )
    input_coordinates.append(
        [(200, 400, 1), (1150, 1200, 1),]
    )
    input_coordinates.append(
        [(250, 350, 1), (1250, 1350, 1),]
    )

    return input_coordinates


@pytest.fixture(scope="module")
def output_coordinates():
    output_coordinates = [
        (100, 200, 1),
        (200, 250, 3),
        (250, 300, 7),
        (300, 350, 6),
        (350, 400, 2),
        (1000, 1100, 1),
        (1150, 1200, 2),
        (1250, 1350, 4),
    ]
    return output_coordinates


@pytest.fixture(scope="module")
def bed_txt_paths(cwas_workspace, input_coordinates):
    return [
        cwas_workspace / f"test{i + 1}.bed"
        for i in range(len(input_coordinates))
    ]


@pytest.fixture(scope="module")
def bed_gz_paths(bed_txt_paths):
    return [Path(str(bed_txt_path) + ".gz") for bed_txt_path in bed_txt_paths]


@pytest.fixture(scope="module", autouse=True)
def setup(cwas_workspace, bed_txt_paths, input_coordinates):
    cwas_workspace.mkdir()
    create_bed_files(bed_txt_paths, input_coordinates)


def create_bed_files(bed_txt_paths, input_coordinates):
    for bed_txt_path, coordinates_per_file in zip(
        bed_txt_paths, input_coordinates
    ):
        create_bed_file(bed_txt_path, bed_entries(coordinates_per_file))
        bed_gz_path = compress_using_bgzip(bed_txt_path)
        _ = index_using_tabix(bed_gz_path)


def bed_entries(coordinates):
    for chrom in chroms():
        for coordinate in coordinates:
            yield (chrom, *coordinate)


def chroms():
    for i in range(NUM_CHROM):
        yield f"chr{i + 1}"


def create_bed_file(bed_file_path, bed_entries):
    print(bed_file_path)
    with bed_file_path.open("w") as bed_file:
        for bed_entry in bed_entries:
            print(*bed_entry, sep="\t", file=bed_file)


@pytest.fixture(scope="module", autouse=True)
def teardown(cwas_workspace):
    yield
    remove_dir(cwas_workspace)


def remove_dir(your_dir):
    for f in your_dir.glob("*"):
        if f.is_dir():
            remove_dir(f)
        else:
            f.unlink()
    your_dir.rmdir()


def test_merge_bed_file_by_chrom(
    cwas_workspace, bed_gz_paths, output_coordinates
):
    chrom = f"chr{random.randint(1, NUM_CHROM)}"
    result_bed_path = cwas_workspace / "result.bed"
    expected = [(chrom, *coordinate) for coordinate in output_coordinates]
    annotation.merge_bed_files_by_chrom(result_bed_path, chrom, bed_gz_paths)

    with result_bed_path.open("r") as result_bed_file:
        for i, line in enumerate(result_bed_file):
            expected_line = "\t".join([str(field) for field in expected[i]])
            assert expected_line == line.strip()

    result_bed_path.unlink()


def test_merge_bed_file_by_chrom_already_exists(cwas_workspace, bed_gz_paths):
    chrom = f"chr{random.randint(1, NUM_CHROM)}"
    result_bed_path = cwas_workspace / "result_empty.bed"
    result_bed_path.touch()

    # Expect the merging process will not run.
    annotation.merge_bed_files_by_chrom(result_bed_path, chrom, bed_gz_paths)

    with result_bed_path.open("r") as result_bed_file:
        assert not result_bed_file.read()  # Expect this file is empty.

    result_bed_path.unlink()


def test_merge_bed_file_by_chrom_force_overwrite(
    cwas_workspace, bed_gz_paths, output_coordinates
):
    chrom = f"chr{random.randint(1, NUM_CHROM)}"
    result_bed_path = cwas_workspace / "result_overwritten.bed"
    result_bed_path.touch()

    with result_bed_path.open("r") as result_bed_file:
        assert not result_bed_file.read()  # Expect this file is empty.

    expected = [(chrom, *coordinate) for coordinate in output_coordinates]
    annotation.merge_bed_files_by_chrom(result_bed_path, chrom, bed_gz_paths, 1)

    with result_bed_path.open("r") as result_bed_file:
        for i, line in enumerate(result_bed_file):
            expected_line = "\t".join([str(field) for field in expected[i]])
            assert expected_line == line.strip()

    result_bed_path.unlink()


def test_merge_bed_files(cwas_workspace, bed_gz_paths, output_coordinates):
    result_bed_path = cwas_workspace / "result.bed"
    bed_file_and_keys = [
        (bed_gz_path, f"BED{i}") for i, bed_gz_path in enumerate(bed_gz_paths)
    ]
    expected_annot_key = "|".join([f"BED{i}" for i in range(len(bed_gz_paths))])
    expected = [
        (f"#ANNOT={expected_annot_key}",),
        ("#chrom", "start", "end", "annot_int"),
    ]

    for i in range(NUM_CHROM):
        chrom = f"chr{i + 1}"
        for coordinate in output_coordinates:
            expected.append((chrom, *coordinate))

    annotation.merge_bed_files(result_bed_path, bed_file_and_keys)
    with result_bed_path.open("r") as result_bed_file:
        for i, line in enumerate(result_bed_file):
            expected_line = "\t".join([str(field) for field in expected[i]])
            assert expected_line == line.strip()

    result_bed_path.unlink()


def test_merge_bed_files_multiprocessing(
    cwas_workspace, bed_gz_paths, output_coordinates
):
    result_bed_path = cwas_workspace / "result.bed"
    bed_file_and_keys = [
        (bed_gz_path, f"BED{i}") for i, bed_gz_path in enumerate(bed_gz_paths)
    ]
    expected_annot_key = "|".join([f"BED{i}" for i in range(len(bed_gz_paths))])
    expected = [
        (f"#ANNOT={expected_annot_key}",),
        ("#chrom", "start", "end", "annot_int"),
    ]

    for i in range(NUM_CHROM):
        chrom = f"chr{i + 1}"
        for coordinate in output_coordinates:
            expected.append((chrom, *coordinate))

    num_proc = random.randint(1, NUM_CHROM)
    annotation.merge_bed_files(result_bed_path, bed_file_and_keys, num_proc)
    with result_bed_path.open("r") as result_bed_file:
        for i, line in enumerate(result_bed_file):
            expected_line = "\t".join([str(field) for field in expected[i]])
            assert expected_line == line.strip()

    result_bed_path.unlink()



"""
Manage CWAS environment variables
"""

from collections import OrderedDict
from pathlib import Path
from typing import Any, Optional

import dotenv


# Ref: https://python-patterns.guide/gang-of-four/singleton/
class Singleton(object):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance


class Env(Singleton):
    def __init__(self, env_path: Path = None):
        if env_path is None:
            if hasattr(self, "path"):
                env_path = self.path
            else:
                env_path = Path.home() / ".cwas_env"

        self.path = env_path
        self.load_env_from_file()

    def get_path(self) -> Path:
        return self.path

    def set_path(self, path: Path):
        """Set the dotenv path"""
        assert path is not None, "The 'path' argument cannot be None."
        self.path = path

    def get_env(self, env_key: str) -> Optional[str]:
        """Return None if the environment value does not exist"""
        return self.env.get(env_key)

    def set_env(self, env_key: str, env_value: Any):
        """Set a new value to the environment variable (key)"""
        self.env[env_key] = str(env_value).strip()

    def reset(self):
        """Make the 'env' attribute empty"""
        self.env = OrderedDict()

    def save(self):
        """Make a new dotenv"""
        with self.path.open("w") as env_f:
            for k, v in self.env.items():
                print(f"{k}={v}", file=env_f)

    def remove_file(self):
        if self.path.exists():
            self.path.unlink()

    def load_env_from_file(self):
        self.env = dotenv.dotenv_values(dotenv_path=self.path)

    def load_env_to_os(self):
        dotenv.load_dotenv(dotenv_path=self.path)



"""
Test the methods in cwas.core.common
"""
import cwas.core.common as common
import numpy as np
import pytest


def test_cmp_two_arr():
    # Empty arrays
    arr1 = np.array([])
    arr2 = np.array([])
    assert common.cmp_two_arr(arr1, arr2)

    # Different lengths
    arr1 = np.array([1, 2, 3, 4, 5])
    arr2 = np.array([3, 4, 5])
    assert not common.cmp_two_arr(arr1, arr2)

    # Same arrays
    arr2 = np.array([1, 2, 3, 4, 5])
    assert common.cmp_two_arr(arr1, arr2)

    # Arrays with the same items but not in the same order
    np.random.shuffle(arr2)
    assert common.cmp_two_arr(arr1, arr2)

    # False if the items of each list are not the same.
    arr2 = np.array([1, 14, 3, 6, 5])
    assert not common.cmp_two_arr(arr1, arr2)


def test_div_dist_num():
    # Raise ValueError if one of any argument is not a positive integer.
    with pytest.raises(ValueError):
        _ = common.div_dist_num(-1, 5)

    with pytest.raises(ValueError):
        _ = common.div_dist_num(5, -1)

    # Raise ValueError if 'num' is less than 'num_group'.
    with pytest.raises(ValueError):
        _ = common.div_dist_num(4, 5)

    num = 5
    num_group = 5
    expected = [1, 1, 1, 1, 1]
    assert expected == common.div_dist_num(num, num_group)

    for i in range(5):
        num += 1
        expected[i] += 1
        assert expected == common.div_dist_num(num, num_group)


def test_chunk_list():
    # ValueError will be raised if the input list is empty.
    with pytest.raises(ValueError):
        _ = common.chunk_list([], 5)

    _list = [1, 2, 3, 4, 5]

    # ValueError will be raised
    # if the number of lists is not a positive integer.
    with pytest.raises(ValueError):
        _ = common.chunk_list(_list, -1)

    with pytest.raises(ValueError):
        _ = common.chunk_list(_list, 0)

    result = common.chunk_list(_list, 1)
    expected = [[1, 2, 3, 4, 5]]
    assert result == expected

    result = common.chunk_list(_list, 2)
    expected = [[1, 2, 3], [4, 5]]
    assert result == expected

    result = common.chunk_list(_list, 3)
    expected = [[1, 2], [3, 4], [5]]
    assert result == expected

    result = common.chunk_list(_list, 4)
    expected = [[1, 2], [3], [4], [5]]
    assert result == expected

    result = common.chunk_list(_list, 5)
    expected = [[1], [2], [3], [4], [5]]
    assert result == expected

    # ValueError is raised if the number of lists exceeds the length of the
    # input list.
    with pytest.raises(ValueError):
        _ = common.chunk_list(_list, 6)


def test_swap_label():
    # Make an input for testing
    label_group_pairs = []

    for n in range(1, 6):
        group_id = f"g{n}"
        label_group_pairs.append((f"{group_id}_l1", group_id))
        label_group_pairs.append((f"{group_id}_l2", group_id))

    np.random.shuffle(label_group_pairs)

    labels = np.array([pair[0] for pair in label_group_pairs])
    group_ids = np.array([pair[1] for pair in label_group_pairs])

    # Basic testing
    swap_labels = common.swap_label(labels, group_ids)
    assert len(swap_labels) == len(group_ids)
    assert np.any(labels != swap_labels)  # Swapping is succeeded.

    # Check if the labels are swapped within their groups.
    for swap_label, group_id in zip(swap_labels, group_ids):
        assert swap_label.startswith(group_id)

    # Test the case where more than 2 labels are in the same group.
    group_id = f"g{np.random.randint(1, 6)}"
    new_label = f"{group_id}_l3"
    labels = np.append(labels, new_label)
    group_ids = np.append(group_ids, group_id)
    np.random.shuffle(labels)
    np.random.shuffle(group_ids)

    with pytest.raises(AssertionError) as e_info:
        assert group_id in str(e_info.value)


def test_int_to_bit_arr():
    assert common.int_to_bit_arr(5, 0).size == 0
    assert common.int_to_bit_arr(5, 3).size == 3
    assert common.int_to_bit_arr(5, 5).size == 5
    assert (common.int_to_bit_arr(1, 3) == np.array([1, 0, 0])).all()
    assert (common.int_to_bit_arr(2, 3) == np.array([0, 1, 0])).all()
    assert (common.int_to_bit_arr(3, 3) == np.array([1, 1, 0])).all()
    assert (common.int_to_bit_arr(4, 3) == np.array([0, 0, 1])).all()
    assert (common.int_to_bit_arr(5, 3) == np.array([1, 0, 1])).all()
    assert (common.int_to_bit_arr(6, 3) == np.array([0, 1, 1])).all()
    assert (common.int_to_bit_arr(7, 3) == np.array([1, 1, 1])).all()


def test_int_to_bit_arr_invalid_args():
    # A negative integer is not allowed as an argument.
    with pytest.raises(ValueError):
        _ = common.int_to_bit_arr(-1, 5)

    with pytest.raises(ValueError):
        _ = common.int_to_bit_arr(5, -1)



"""
Test cwas.preparation
"""
import pytest
import yaml
from cwas.annotation import Annotation
from cwas.env import Env


class AnnotationMock(Annotation):
    """Mocking the Annotation class"""

    def run(self):
        pass


@pytest.fixture(scope="module")
def vcf_path(cwas_workspace):
    return cwas_workspace / "test_target.vcf"


@pytest.fixture(scope="module", autouse=True)
def setup(cwas_workspace, annotation_dir, vcf_path):
    cwas_workspace.mkdir()
    set_env(cwas_workspace, annotation_dir)
    create_vcf_file(vcf_path)
    create_bw_key_yaml(annotation_dir / "bw_key.yaml")


@pytest.fixture(scope="module", autouse=True)
def teardown(cwas_workspace):
    yield
    env = Env()
    env.reset()
    env.remove_file()
    for f in cwas_workspace.glob("*"):
        f.unlink()
    cwas_workspace.rmdir()


def set_env(cwas_workspace, annotation_dir):
    env = Env()
    env.set_env("CWAS_WORKSPACE", cwas_workspace)
    env.set_env("VEP", "VEP")
    env.set_env("ANNOTATION_DATA", annotation_dir)
    env.set_env("ANNOTATION_BW_KEY", annotation_dir / "bw_key.yaml")
    env.set_env("MERGED_BED", cwas_workspace / "merged.bed.gz")
    env.set_env("MERGED_BED_INDEX", cwas_workspace / "merged.bed.gz.tbi")
    env.save()


def create_vcf_file(vcf_path):
    vcf_header = (
        "#CHROM",
        "POS",
        "ID",
        "REF",
        "ALT",
        "QUAL",
        "FILTER",
        "INFO",
    )
    # TODO: Add VCF entries (variants)
    with vcf_path.open("w") as vcf_file:
        print(*vcf_header, sep="\t", file=vcf_file)


@pytest.fixture(scope="module")
def required_args(vcf_path):
    return ["-v", str(vcf_path)]


def test_parse_args(required_args, vcf_path):
    inst = AnnotationMock.get_instance(required_args)
    assert getattr(inst, "vcf_path") == vcf_path


def test_parse_args_without_required_arg():
    args = []
    with pytest.raises(SystemExit):
        AnnotationMock.get_instance(args)




from collections import deque
import multiprocessing as mp ## test multiprocessing
import pysam
import itertools
import gzip

class annotate:
    def __init__(self, in_vcf_gz_path: str,
                 out_vcf_path: str, annot_bed_path: str, num_proc: str) -> None:
        self._in_vcf_gz_path = in_vcf_gz_path
        self._out_vcf_path = out_vcf_path
        self._annot_bed_path = annot_bed_path
        self._num_proc = num_proc if num_proc < 23 else 22

    @property
    def in_vcf_gz_path(self) -> str:
        return self._in_vcf_gz_path
    @property
    def out_vcf_path(self) -> str:
        return self._out_vcf_path
    @property
    def annot_bed_path(self) -> str:
        return self._annot_bed_path
    @property
    def num_proc(self) -> int:
        return self._num_proc

    def bed_custom_annotate(self):
        with pysam.TabixFile(self.in_vcf_gz_path) as in_vcf_file, pysam.TabixFile(
            self.annot_bed_path
        ) as annot_bed_file, gzip.open(self.out_vcf_path, "wt") as out_vcf_file:
            # Make and write headers
            vcf_headers = in_vcf_file.header
            annot_key_str = annot_bed_file.header[0].split("=")[1]
            annot_info_header = f"##INFO=<ID=ANNOT,Key={annot_key_str}>"
            vcf_headers.append(annot_info_header)
            vcf_headers[-1], vcf_headers[-2] = (
                vcf_headers[-2],
                vcf_headers[-1],
            )  # Swap
            
            # Write VCF header
            out_vcf_file.write("\n".join(vcf_headers))

            # Annotate by the input BED file
            vcf_chroms = in_vcf_file.contigs
            self._num_proc = self._num_proc if len(vcf_chroms) > self._num_proc else len(vcf_chroms)

            p = mp.Pool(self._num_proc)

            annot_vcf = p.map(self.chr_annotate, vcf_chroms)
            annot_vcf = itertools.chain.from_iterable(annot_vcf)
            
            p.close()
            p.join()
            
            # Write VCF annotated
            out_vcf_file.write("\n")
            out_vcf_file.write("\n".join(annot_vcf))
            out_vcf_file.close()

    def chr_annotate(self, chrom):
        input_vcf = pysam.TabixFile(self.in_vcf_gz_path)
        annot_bed = pysam.TabixFile(self.annot_bed_path)

        var_iter = input_vcf.fetch(chrom, parser=pysam.asTuple())
        bed_iter = annot_bed.fetch(chrom, parser=pysam.asTuple())
        bed_memory = deque()
        variant = next(var_iter, None) # var_iter의 마지막 -> None
        variant_annot = []

        while variant is not None:
            var_pos = int(variant[1]) - 1  # 1-based -> 0-based
            var_ref = variant[3]
            var_alt = variant[4]

            # Determine a search region for BED coordinates
            if len(var_ref) == 1:  # Insertion or substitution
                region_start = var_pos
                region_end = (
                var_pos + 2 if len(var_alt) > 1 else var_pos + 1
                )
            else:  # Deletion
                region_start = var_pos + 1
                region_end = region_start + len(var_ref) - 1

            # Get an annotation integer
            annot_int = 0
            stop_bed_iter = False

            # 1. Check the memory of previously checked BED coordinates
            while (
                len(bed_memory) > 0
                and int(bed_memory[0][2]) <= region_start
            ):
                # Remove non-overlapped coordinates
                bed_memory.popleft()

            for bed in bed_memory:
                if int(bed[1]) < region_end:  # Overlap
                    annot_int |= int(bed[3])
                else:
                    stop_bed_iter = True
                    break

            # 2. Continuously iterate over the BED coordinates and check
            if not stop_bed_iter:
                bed = next(bed_iter, None)

                while bed is not None:
                    bed_start = int(bed[1])
                    bed_end = int(bed[2])

                    if region_start < bed_end:
                        bed_memory.append(bed)
                        if bed_start < region_end:  # Overlap
                            annot_int |= int(bed[3])
                        else:
                            break

                    bed = next(bed_iter, None)

            variant_annot.append(str(variant) + f";ANNOT={annot_int}")

            #print(str(variant) + f";ANNOT={annot_int}", file=out_vcf)
            variant = next(var_iter, None)

        return variant_annot


"""
Create configuration files for CWAS
"""
from pathlib import Path

import yaml
from cwas.core.configuration.settings import (
    get_default_domains,
    get_domain_types,
    get_redundant_domain_pairs,
)
from cwas.utils.log import print_err


def create_annotation_key(
    out_file_path: Path, annotation_dir: Path, file_type: str
):
    assert file_type == "bed" or file_type == "bw"

    glob_key = "*.bed.gz" if file_type == "bed" else "*.bw"
    file_ext_len = len(glob_key) - 1  # Ignore '*'
    filenames = [
        str(file_path.name) for file_path in annotation_dir.glob(glob_key)
    ]
    annot_key_dict = {
        filename: filename[:-file_ext_len].replace(".", "_")
        for filename in filenames
    }
    _save_as_yaml(out_file_path, annot_key_dict)

def create_category_domain_list(
    domain_list_path: Path,
    bed_key_path: Path,
    gene_mat_path: Path,
):
    bed_key_dict = _load_yaml_file(bed_key_path)
    
    cons_domains = bed_key_dict['functional_score'].values()
    region_domains = bed_key_dict['functional_annotation'].values()
    
    with gene_mat_path.open("r") as gene_mat_f:
        header = gene_mat_f.readline()
        columns = header.strip().split()
        gene_list_domains = columns[2:]  # Remove 'gene_id' and 'gene_name'

    domains = get_default_domains()
    domains["functional_annotation"] += region_domains
    domains["functional_score"] += cons_domains
    domains["gene_set"] += gene_list_domains

    _save_as_yaml(domain_list_path, domains)


def create_redundant_category_table(category_table_path: Path):
    domain_types = get_domain_types()
    table_row_template = {domain_type: "*" for domain_type in domain_types}

    with category_table_path.open("w") as out_f:
        print(*domain_types, sep="\t", file=out_f)
        redundant_domain_pairs = get_redundant_domain_pairs()
        for domain_type_pair, domain_pair_set in redundant_domain_pairs.items():
            for domain_pair in domain_pair_set:
                table_row = dict(table_row_template)
                for i, domain_type in enumerate(domain_type_pair):
                    table_row[domain_type] = domain_pair[i]
                print(
                    *[table_row[domain_type] for domain_type in domain_types],
                    sep="\t",
                    file=out_f,
                )


def _load_yaml_file(yaml_file_path: Path):
    try:
        with yaml_file_path.open("r") as in_yaml_f:
            yaml_data = yaml.safe_load(in_yaml_f)
    except yaml.YAMLError:
        print_err(f'"{yaml_file_path}" is not in a proper YAML format.')
        raise
    return yaml_data


def _save_as_yaml(yaml_file_path: Path, data):
    with yaml_file_path.open("w") as out_yaml_f:
        yaml.safe_dump(data, out_yaml_f)



import numpy as np
from scipy.stats import norm
import matplotlib.pyplot as plt
import os

from cwas.burden_test import BurdenTest
from cwas.core.burden_test.binomial import binom_one_tail, binom_two_tail
from cwas.utils.log import print_progress
from scipy.stats import binomtest
from decimal import Decimal


class BinomialTest(BurdenTest):
    @property
    def binom_p(self) -> float:
        return self.case_cnt / (self.case_cnt + self.ctrl_cnt)
    
    @property
    def count_thres(self) -> int:
        if self._count_thres is None:
            m = 1
            while True:
                p_value = binomtest(m-1, m, self.binom_p, alternative='greater').pvalue
                if p_value < 0.05:
                    return m
                m += 1

    def run_burden_test(self):
        print_progress("Run binomial test")
        if self.use_n_carrier:
            n1 = self.case_carrier_cnt
            n2 = self.ctrl_carrier_cnt
        else:
            n1 = self.case_variant_cnt.round()
            n2 = self.ctrl_variant_cnt.round()
        self._result["P"] = np.vectorize(binom_two_tail)(
            n1,
            n2,
            self.binom_p,
        )

        # Add the pseudocount(1) in order to avoid p-values of one
        self._result["P_1side"] = np.vectorize(binom_one_tail)(
            n1 + 1,
            n2 + 1,
            self.binom_p,
        )
        self._result["Z_1side"] = norm.ppf(1 - self._result["P_1side"].values)

        self._draw_volcano_plot()

    def _draw_volcano_plot(self):
        burden_res = self._result.copy()
        burden_res['log2_RR'] = burden_res['Relative_Risk'].apply(lambda x: np.log2(x))
        burden_res['-log_P'] = burden_res['P'].apply(lambda x: -np.log10(x))
        
        print_progress(f"Volcano plot will display categories with at least {self.count_thres} counts")
        selected_categories = self._raw_counts[self._raw_counts['Raw_counts'] >= self.count_thres].index.tolist()        
        burden_res = burden_res[burden_res.index.isin(selected_categories)]
        
        threshold = -np.log10(0.05)
        if self.eff_test:
            eff_threshold = -np.log10(0.05/self.eff_test)
        max_rr = max(burden_res.loc[burden_res.log2_RR!=np.inf, 'log2_RR'])
        min_rr = min(burden_res.loc[burden_res.log2_RR!=-np.inf, 'log2_RR'])
        max_val = max(abs(max_rr),abs(min_rr))
        max_x = np.trunc(max_val) + 2

        xticks = [int(x) for x in np.arange(-max_x, max_x+1, 2)]
        xlabels = xticks.copy()
        xlabels[0] = '-Inf'
        xlabels[-1] = 'Inf'
        yticks = [int(x) for x in np.arange(0, max(burden_res['-log_P']), 2)]
        ylabels = yticks.copy()

        def replace_inf(x, v):
            if x == np.inf:
                return v
            elif x == -np.inf:
                return -v
            else:
                return x

        if self.tag != None:
            tags = self.tag.strip().split(",")
            
            for t in tags:
                fig, axes = plt.subplots(figsize=(self.plot_size, self.plot_size))

                plt.title(self.plot_title, fontsize=self.font_size, loc='left', pad=5)
                axes.vlines(x=0, ymin=-0.5, ymax=max(burden_res['-log_P'])+0.5, linestyles='-', color='lightgray', linewidth=1.25, zorder=1)
                axes.scatter(x=burden_res['log2_RR'].apply(lambda x: replace_inf(x, max_x)), y=burden_res['-log_P'],
                            marker='o', color='silver', s=self.marker_size, label='Others', edgecolor='black', linewidth=0.5, zorder=2)
                axes.scatter(x=burden_res.loc[(burden_res.index.str.contains(t))&(burden_res['-log_P']>threshold), 'log2_RR'], 
                             y=burden_res.loc[(burden_res.index.str.contains(t))&(burden_res['-log_P']>threshold), '-log_P'],
                             marker='o', label=t, facecolor='#3d62a1', s=self.marker_size, alpha=.7, edgecolor='black', linewidth=0.5, zorder=3)
                axes.hlines(y=threshold, xmin=-(max_x+1), xmax=max_x+1, linestyles='--', linewidth=1.25, color='black')
                axes.text(-(max_x+1)+0.1, threshold+0.1, 'P=0.05', size=self.font_size*0.85, color='black')
                if self.eff_test:
                    axes.hlines(y=eff_threshold, xmin=-(max_x+1), xmax=max_x+1, linestyles='--', linewidth=1.25, color='red')
                    axes.text(-(max_x+1)+0.1, eff_threshold+0.1, ''.join(['P=', str('%.2E' % Decimal(0.05/self.eff_test)),', eff_num=', format(self.eff_test, ',')]), size=self.font_size*0.85, color='red')
                plt.ylim(-0.5, max(burden_res['-log_P'])+0.5)
                plt.xlim(-(max_x+1), (max_x+1))
                plt.xlabel('Relative Risk ($log_{2}$)', size=self.font_size)
                plt.ylabel("P ($-log_{10}$)", size=self.font_size)
                axes.set_xticks(xticks)
                axes.set_xticklabels(xlabels, fontsize=self.font_size)
                axes.set_yticks(yticks)
                axes.set_yticklabels(ylabels, fontsize=self.font_size)
                axes.legend(markerscale=self.font_size*0.15, fontsize=self.font_size*0.85)
                axes.spines['bottom'].set_linewidth(1.25)
                axes.spines['left'].set_linewidth(1.25)
                axes.tick_params(width=1.25)
                axes.spines['top'].set_visible(False)
                axes.spines['right'].set_visible(False)
                
                output = self.result_path.name.replace(".txt", f'.{t}.volcano_plot.pdf')
                output_path = os.path.join(self.output_dir_path, output)
                
                plt.tight_layout()
                plt.savefig(output_path, bbox_inches='tight')
                print_progress("Save the result to the volcano plot file {}".format(output_path))
        else:
            fig, axes = plt.subplots(figsize=(self.plot_size, self.plot_size))

            plt.title(self.plot_title, fontsize=self.font_size, loc='left', pad=5)
            axes.vlines(x=0, ymin=-0.5, ymax=max(burden_res['-log_P'])+0.5, linestyles='-', color='lightgray', linewidth=1.25, zorder=1)
            axes.scatter(x=burden_res['log2_RR'].apply(lambda x: replace_inf(x, max_x)), y=burden_res['-log_P'],
                        marker='o', color='silver', s=self.marker_size, edgecolor='black', linewidth=0.5, zorder=2)
            axes.hlines(y=threshold, xmin=-(max_x+1), xmax=max_x+1, linestyles='--', linewidth=1.25, color='black')
            axes.text(-(max_x+1)+0.1, threshold+0.1, 'P=0.05', size=self.font_size*0.85, color='black')
            if self.eff_test:
                axes.hlines(y=eff_threshold, xmin=-(max_x+1), xmax=max_x+1, linestyles='--', linewidth=1.25, color='red')
                axes.text(-(max_x+1)+0.1, eff_threshold+0.1, ''.join(['P=', str('%.2E' % Decimal(0.05/self.eff_test)),', eff_num=', format(self.eff_test, ',')]), size=self.font_size*0.85, color='red')
            plt.ylim(-0.5, max(burden_res['-log_P'])+0.5)
            plt.xlim(-(max_x+1), (max_x+1))
            plt.xlabel('Relative Risk ($log_{2}$)', size=self.font_size)
            plt.ylabel("P ($-log_{10}$)", size=self.font_size)
            axes.set_xticks(xticks)
            axes.set_xticklabels(xlabels, fontsize=self.font_size)
            axes.set_yticks(yticks)
            axes.set_yticklabels(ylabels, fontsize=self.font_size)
            axes.spines['bottom'].set_linewidth(1.25)
            axes.spines['left'].set_linewidth(1.25)
            axes.tick_params(width=1.25)
            axes.spines['top'].set_visible(False)
            axes.spines['right'].set_visible(False)

            output = self.result_path.name.replace(".txt", f'.volcano_plot.pdf')
            output_path = os.path.join(self.output_dir_path, output)

            plt.tight_layout()
            plt.savefig(output_path, bbox_inches='tight')
            print_progress("Save the result to the volcano plot file {}".format(output_path))



"""
Functions for printing logs for this project.
"""
import sys
from datetime import datetime


def print_log(log_type: str, log_msg: str, print_time: bool = False):
    log_header = (
        f"[{_get_curr_time()}, {log_type}]" if print_time else f"[{log_type}]"
    )
    _eprint(f"{log_header} {log_msg}")


def print_arg(arg_name: str, arg_val: str):
    msg = f"{arg_name}: {arg_val}"
    print_log("ARG", msg)


def print_progress(msg: str):
    print_log("PROGRESS", msg, True)


def print_warn(msg: str):
    print_log("WARNING", msg, True)


def print_err(msg: str):
    print_log("ERROR", msg, True)


# Ref:
# https://stackoverflow.com/questions/5574702/how-to-print-to-stderr-in-python
def _eprint(*args, **kwargs):
    """
    print to the stderr
    """
    print(*args, file=sys.stderr, **kwargs)


def _get_curr_time() -> str:
    now = datetime.now()
    curr_time = now.strftime("%H:%M:%S %m/%d/%y")
    return curr_time



"""
Tests the situation when we call CWAS Start again
"""
from pathlib import Path

import pytest
from cwas.env import Env
from cwas.start import Start


@pytest.fixture(scope="module")
def args(cwas_workspace: Path) -> list:
    return ["-w", str(cwas_workspace)]


@pytest.fixture(scope="module")
def config() -> dict:
    test_config = {"TEST": "HELLO"}
    return test_config


@pytest.fixture(scope="module")
def env() -> dict:
    test_env = {"FAKE_ENV": "HI"}
    return test_env


@pytest.fixture(scope="module")
def create_config_early(cwas_workspace: Path, config: dict):
    cwas_workspace.mkdir()
    cwas_config_path = cwas_workspace / "configuration.txt"

    with cwas_config_path.open("w") as config_file:
        for config_key, config_value in config.items():
            print(f"{config_key}={config_value}", file=config_file)


@pytest.fixture(scope="module")
def create_env_early(cwas_env_path: Path, env: dict):
    with cwas_env_path.open("w") as env_file:
        for env_key, env_value in env.items():
            print(f"{env_key}={env_value}", file=env_file)


@pytest.fixture(scope="module", autouse=True)
def setup(args: list, create_config_early, create_env_early):
    inst = Start.get_instance(args)
    inst.run()


@pytest.fixture(scope="module", autouse=True)
def teardown(cwas_workspace: Path, cwas_env_path: Path):
    yield
    cwas_env_path.unlink()
    for f in cwas_workspace.glob("*"):
        f.unlink()
    cwas_workspace.rmdir()


def test_cwas_config(cwas_workspace: Path, config: dict):
    cwas_config_path = cwas_workspace / "configuration.txt"
    actual_config = get_config_from_file(cwas_config_path)
    assert actual_config.items() == config.items()


def test_cwas_env_path(cwas_env_path: Path, env: dict):
    actual_env = get_config_from_file(cwas_env_path)
    expected_key_set = {
        "CWAS_WORKSPACE",
        "FAKE_ENV",
    }
    assert actual_env.items() >= env.items()
    assert set(actual_env.keys()) == expected_key_set


def get_config_from_file(config_path: Path) -> dict:
    config_from_file = {}
    with config_path.open() as config_file:
        for line in config_file:
            k, v = line.strip().split("=")
            config_from_file[k] = v
    return config_from_file



"""
Common algorithms for CWAS
"""
import numpy as np


def cmp_two_arr(array1: np.ndarray, array2: np.ndarray) -> bool:
    """ Return True if two arrays have the same items regardless of the order.
    Otherwise, it returns False.
    """
    if len(array1) != len(array2):
        return False

    array1_item_set = set(array1)

    for item in array2:
        if item not in array1_item_set:
            return False

    return True


def div_dist_num(num: int, num_group: int) -> list:
    """ Divide and distribute the number to each group almost equally. """
    if num <= 0 or num_group <= 0:
        raise ValueError("Only positive integers are accepted as arguments.")

    if num < num_group:
        raise ValueError(
            'The argument "num" must be larger than the argument "num_group".'
        )

    num_per_groups = []
    num_per_group = num // num_group
    remain_num = num % num_group

    for _ in range(num_group):
        if remain_num == 0:
            num_per_groups.append(num_per_group)
        else:
            num_per_groups.append(num_per_group + 1)
            remain_num -= 1

    return num_per_groups


def chunk_list(_list: list, num_chunk: int) -> list:
    """ Split the input list into multiple chunks """
    if len(_list) == 0:
        raise ValueError(
            "This function does not accept an empty list as an argument."
        )

    if num_chunk <= 0:
        raise ValueError("The number of chunks must be a positive integer.")

    if len(_list) < num_chunk:
        raise ValueError(
            "The length of the input list must be larger "
            "than the number of chunks."
        )

    chunks = []
    chunk_lens = div_dist_num(len(_list), num_chunk)

    # Make chunks
    i = 0
    for chunk_len in chunk_lens:
        chunk = _list[i : i + chunk_len]
        chunks.append(chunk)
        i += chunk_len

    return chunks


def swap_label(labels: np.ndarray, group_ids: np.ndarray) -> np.ndarray:
    """ Randomly swap labels (case or control) in each group
    and return a list of swapped labels.

    :param labels: Array of labels
    :param group_ids: Array of group IDs corresponding to each label
    :return: Swapped labels
    """
    # Key: a group ID, Value: No. times the key is referred
    group_to_hit_cnt = {group_id: 0 for group_id in group_ids}
    # Key: A group, Value: The index of a label firstly matched with the group
    group_to_idx = {}
    swap_labels = np.copy(labels)

    # Make an array for random swapping
    num_group = len(group_to_hit_cnt.keys())
    do_swaps = np.random.binomial(1, 0.5, size=num_group)
    group_idx = 0

    for i, label in enumerate(labels):
        group_id = group_ids[i]
        group_hit_cnt = group_to_hit_cnt.get(group_id, 0)
        assert (
            group_hit_cnt == 0 or group_hit_cnt == 1
        ), f'Too many labels (more than 2) in a group "{group_id}".'

        if group_hit_cnt == 0:
            group_to_hit_cnt[group_id] = 1
            group_to_idx[group_id] = i
        else:
            group_to_hit_cnt[group_id] += 1
            prev_idx = group_to_idx[group_id]

            # Swap
            do_swap = do_swaps[group_idx]
            if do_swap:
                swap_labels[i] = labels[prev_idx]
                swap_labels[prev_idx] = label

            group_idx += 1

    return swap_labels


def int_to_bit_arr(n0: float, bit_arr_len: int) -> np.ndarray:
    n = int(n0)
    if n < 0 or bit_arr_len < 0:
        raise ValueError("A negative integer argument does not allowed.")

    bit_arr = np.zeros(bit_arr_len)

    for i in range(bit_arr_len):
        bit = n % 2
        bit_arr[i] += bit
        n >>= 1

        if n == 0:
            break

    return bit_arr



"""
Test cwas.preparation
"""
import random
from multiprocessing import cpu_count
from pathlib import Path
from typing import Tuple

import pytest
from cwas.env import Env
from cwas.preparation import Preparation


class PreparationMock(Preparation):
    """Mocking the Preparation class"""

    def _prepare_annotation(self) -> Tuple[Path, Path]:
        """This step is supposed to be tested in 
        tests/core/preparation/test_annotation.py
        """
        merged_bed_path = self.workspace / "merged.bed.gz"
        merged_bed_idx_path = self.workspace / "merged.bed.gz.tbi"
        return merged_bed_path, merged_bed_idx_path


@pytest.fixture(scope="module", autouse=True)
def setup(cwas_workspace: Path, annotation_dir: Path):
    cwas_workspace.mkdir()
    set_env(cwas_workspace, annotation_dir)


@pytest.fixture(scope="module", autouse=True)
def teardown(cwas_workspace: Path):
    yield
    reset_env()
    remove_workspace(cwas_workspace)


def set_env(cwas_workspace: Path, annotation_dir: Path):
    env = Env()
    env.set_env("CWAS_WORKSPACE", cwas_workspace)
    env.set_env("ANNOTATION_DATA", annotation_dir)
    env.set_env("ANNOTATION_BED_KEY", cwas_workspace / "bed_key.yaml")
    env.save()


def reset_env():
    env = Env()
    env.reset()
    env.remove_file()


def remove_workspace(cwas_workspace: Path):
    for f in cwas_workspace.glob("*"):
        f.unlink()
    cwas_workspace.rmdir()


def test_default_args():
    inst = PreparationMock.get_instance()
    assert getattr(inst, "num_proc") == 1
    assert getattr(inst, "force_overwrite") == 0


def test_parse_args():
    cpu = random.choice(range(1, cpu_count() + 1))
    args = ["-p", str(cpu), "--force_overwrite"]
    inst = PreparationMock.get_instance(args)
    assert getattr(inst, "num_proc") == cpu
    assert getattr(inst, "force_overwrite") == 1

    args = ["-p", str(cpu)]
    inst = PreparationMock.get_instance(args)
    assert getattr(inst, "num_proc") == cpu
    assert getattr(inst, "force_overwrite") == 0

    args = ["--force_overwrite"]
    inst = PreparationMock.get_instance(args)
    assert getattr(inst, "num_proc") == 1
    assert getattr(inst, "force_overwrite") == 1


def test_parse_args_value_error():
    cpu = cpu_count() + 1
    args = ["-p", str(cpu)]
    with pytest.raises(ValueError):
        PreparationMock.get_instance(args)

    args = ["-p", "0"]
    with pytest.raises(ValueError):
        PreparationMock.get_instance(args)

    args = ["-p", "-1"]
    with pytest.raises(ValueError):
        PreparationMock.get_instance(args)


def test_env_after_run_preparation():
    inst = PreparationMock.get_instance()
    inst.run()
    assert inst.get_env("MERGED_BED")
    assert inst.get_env("MERGED_BED_INDEX")


def test_run_without_configuration():
    inst = PreparationMock.get_instance()
    Env().reset()
    with pytest.raises(RuntimeError):
        inst.run()



from __future__ import annotations

import argparse
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional

from cwas.env import Env


class Runnable(ABC):
    def __init__(self, args: Optional[argparse.Namespace] = None):
        self.env = Env()
        self._args = args
        self._print_args(args)
        self._check_args_validity(args)

    @property
    def args(self) -> Optional[argparse.Namespace]:
        return argparse.Namespace(**vars(self._args)) if self._args else None

    def set_env_path(self, path: Path):
        self.env.set_path(path)

    def set_env(self, env_key: str, env_value: Any):
        self.env.set_env(env_key, env_value)

    def get_env(self, env_key: str) -> Optional[str]:
        return self.env.get_env(env_key)

    def save_env(self):
        self.env.save()

    @staticmethod
    def _create_arg_parser() -> argparse.ArgumentParser:
        return argparse.ArgumentParser()

    @staticmethod
    def _print_args(args: argparse.Namespace):
        pass

    @staticmethod
    def _check_args_validity(args: argparse.Namespace):
        pass

    @abstractmethod
    def run(self):
        pass



"""
CWAS Annotation Step

This step annotate user's VCF file using annotation data specified
in the CWAS configuration step. This step mainly uses
Variant Effect Predictor (VEP) to annotate user's VCF file.
"""
import argparse
from pathlib import Path
from typing import Optional

from cwas.core.annotation.bed import annotate as _annotate_using_bed
from cwas.core.annotation.vep import VepCmdGenerator
from cwas.runnable import Runnable
from cwas.utils.check import check_is_file
from cwas.utils.check import check_is_dir
from cwas.utils.check import check_num_proc
from cwas.utils.cmd import CmdExecutor, compress_using_bgzip, index_using_tabix
from cwas.utils.log import print_arg, print_log, print_progress

import multiprocessing as mp
from functools import partial
from pysam import TabixFile


class Annotation(Runnable):
    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self._vcf_path = None

    @staticmethod
    def _print_args(args: argparse.Namespace):
        print_arg("Target VCF file", args.vcf_path)
        print_arg("Output directory", args.output_dir_path)
        print_arg("Number of worker processes", args.num_proc)

    @staticmethod
    def _check_args_validity(args: argparse.Namespace):
        check_is_file(args.vcf_path)
        check_is_dir(args.output_dir_path)
        check_num_proc(args.num_proc)

    @property
    def vcf_path(self):
        if self._vcf_path is None:
            if (self.args.num_proc > 1) and (self.args.vcf_path.suffix != '.gz'):
                vcf_gz = compress_using_bgzip(self.args.vcf_path)
                index_using_tabix(vcf_gz)
                self._vcf_path = vcf_gz
            else:
                self._vcf_path = self.args.vcf_path.resolve()
        
        return self._vcf_path
    
    @property
    def num_proc(self):
        return self.args.num_proc

    @property
    def output_dir_path(self):
        return self.args.output_dir_path.resolve()

    @property
    def vep_cmd(self):
        vep_cmd_generator = VepCmdGenerator(
            self.get_env("VEP"), self.get_env("VEP_CACHE_DIR"),
            self.get_env("VEP_CONSERVATION_FILE"), self.get_env("VEP_LOFTEE"), 
            self.get_env("VEP_HUMAN_ANCESTOR_FA"), self.get_env("VEP_GERP_BIGWIG"), 
            self.get_env("VEP_MIS_DB"), self.get_env("VEP_MIS_INFO_KEY"),
            str(self.vcf_path), str(self.num_proc),
        )
        vep_cmd_generator.output_vcf_path = self.vep_output_vcf_path
        return vep_cmd_generator.cmd

    @property
    def vep_output_vcf_path(self):
        if self.vcf_path.suffix == '.gz':
            return (
                f"{self.output_dir_path}/"
                f"{self.vcf_path.stem.replace('.vcf', '.vep.vcf')}"
            )
        else:
            return (
                f"{self.output_dir_path}/"
                f"{self.vcf_path.name.replace('.vcf', '.vep.vcf')}"
            )

    @property
    def vep_output_vcf_gz_path(self):
        return self.vep_output_vcf_path.replace(".vcf", ".vcf.gz")

    @property
    def annotated_vcf_path(self):
        return self.vep_output_vcf_gz_path.replace('.vep.vcf.gz', '.annotated.vcf.gz')

    def run(self):
        self.annotate_using_vep()
        self.process_vep_vcf()
        self.annotate_using_bed()
        self.update_env()
        print_progress("Done")

    def annotate_using_vep(self):      
        print_progress("Annotation via VEP")
        if (
            Path(self.vep_output_vcf_path).is_file()
            or Path(self.vep_output_vcf_gz_path).is_file()
        ):
            print_log(
                "NOTICE",
                "You have already done the VEP annotations.",
                True,
            )
            return
        
        if self.num_proc == 1:
            vep_bin, *vep_args = self.vep_cmd
            CmdExecutor(vep_bin, vep_args).execute_raising_err()
        else:
            vep_bin, _, _, _, _, *vep_args = self.vep_cmd
            
            print_progress("For multiprocessing, the input VCF should be indexed")
            
            chroms = self.fetch_chromosomes()
            
            multi_inputs = []
            args_list = []
            tmp_output_list = []
            for i in chroms:
                multi_inputs.append(' '.join(['tabix -h', str(self.vcf_path), i, '|']))
                replace_name = '.' + i + '.vep.vcf'
                tmp_output_vcf_path = self.vep_output_vcf_path.replace(".vep.vcf", replace_name)
                args_list.append(['-o', tmp_output_vcf_path, *vep_args])
                tmp_output_list.append(tmp_output_vcf_path)
                
            print_progress(' '.join(["Input VCF has", str(len(chroms)), "number of chromosomes"]))
            
            num_processes = self.num_proc if self.num_proc < len(chroms) else len(chroms)
            
            _run_multiple_vep = partial(self.execute_CMD_mp,
                                        shell = True)
            
            # Create a multiprocessing pool
            pool = mp.Pool(processes=num_processes)
            
            # Use starmap to pass args_list and multi_inputs (To keep vep_bin as the first element, the code will repeat it during mapping)
            pool.starmap(_run_multiple_vep, zip([vep_bin for _ in range(len(chroms))], args_list, multi_inputs))

            # Close the pool and wait for all processes to finish
            pool.close()
            pool.join()
            
            print_progress("Merge and sort output files into a single file")
            args_header = ["'^#'", tmp_output_list[0], ">", self.vep_output_vcf_path]
            CmdExecutor("grep", args_header, shell=True).execute_raising_err()
            args_merge = ["-k1,1V", "-k2,2n", '>>', self.vep_output_vcf_path]
            CmdExecutor("sort", args_merge,
                        multi_input = ' '.join(["grep", "--no-filename", "-v", "'^#'", *tmp_output_list, '|']),
                        shell=True).execute_raising_err()
            print_progress("Remove temporary outputs")
            args_remove = [*tmp_output_list]
            CmdExecutor("rm", args_remove).execute_raising_err()

    def execute_CMD_mp(self, bin: str, args: list = [], multi_input: Optional[str] = None, shell: bool = False):
        return CmdExecutor(bin = bin, args = args, multi_input = multi_input, shell=shell).execute_raising_err()
    
    def fetch_chromosomes(self):
        chromosomes = ['chr' + str(i) for i in range(1, 23)] + ['chrX', 'chrY']
        chr_list = []
        for chromosome in chromosomes:
            try:
                vcf_reader = TabixFile(str(self.vcf_path))
                vcf_reader.fetch(chromosome)
                chr_list.append(chromosome)
            except StopIteration:
                pass
            except Exception:
                pass
        return chr_list

    def process_vep_vcf(self):
        print_progress("Compress the VEP output using bgzip")
        vcf_gz_path = compress_using_bgzip(self.vep_output_vcf_path)
        print_progress("Create an index of the VEP output using tabix")
        index_using_tabix(vcf_gz_path)

    def annotate_using_bed(self):
        print_progress("BED custom annotation")
        if Path(self.annotated_vcf_path).is_file():
            print_log(
                "NOTICE",
                "You have already done the BED custom annotation.",
                True,
            )
            return
        
        annotate_vcf = _annotate_using_bed(
            self.vep_output_vcf_gz_path,
            self.annotated_vcf_path,
            self.get_env("MERGED_BED"),
            self.num_proc,
        )

        annotate_vcf.bed_custom_annotate()

    def update_env(self):
        self.set_env("ANNOTATED_VCF", self.annotated_vcf_path)
        self.save_env()


import argparse
from pathlib import Path
from typing import Tuple

import yaml

import cwas.utils.log as log
from cwas.core.preparation.annotation import merge_bed_files
from cwas.runnable import Runnable
from cwas.utils.check import check_num_proc
from cwas.utils.cmd import compress_using_bgzip, index_using_tabix


class Preparation(Runnable):
    def __init__(self, args: argparse.Namespace):
        super().__init__(args)

    @staticmethod
    def _print_args(args: argparse.Namespace):
        log.print_arg("No. Processes for this step", args.num_proc)
        log.print_arg(
            "Force to overwrite the result",
            "Y" if args.force_overwrite else "N",
        )

    @staticmethod
    def _check_args_validity(args: argparse.Namespace):
        check_num_proc(args.num_proc)

    @property
    def num_proc(self):
        return self.args.num_proc

    @property
    def force_overwrite(self):
        return bool(self.args.force_overwrite)

    def run(self):
        self._load_env()
        bed_gz_path, bed_idx_path = self._prepare_annotation()
        self._save_as_env(bed_gz_path, bed_idx_path)

    def _load_env(self):
        """Load environment variables to attributes"""
        try:
            self.workspace = Path(self.get_env("CWAS_WORKSPACE"))
            self.annot_data_dir = Path(self.get_env("ANNOTATION_DATA"))
            self.bed_key_list_path = self.workspace / self.get_env(
                "ANNOTATION_BED_KEY"
            )
        except TypeError:
            raise RuntimeError(
                "Failed to get one of CWAS environment variable."
                " Maybe you omitted to run Configuration step."
            )

    def _prepare_annotation(self) -> Tuple[Path, Path]:
        log.print_progress(
            "Data preprocessing to prepare CWAS annotation step"
        )

        with self.bed_key_list_path.open() as bed_key_list_file:
            bed_key_list = yaml.safe_load(bed_key_list_file)
        
        # Merge two dictionaries
        bed_key_list = bed_key_list['functional_score'] | bed_key_list['functional_annotation']

        bed_file_and_keys = []
        for bed_filename, bed_key in bed_key_list.items():
            bed_file_path = self.annot_data_dir / bed_filename
            bed_file_and_keys.append((bed_file_path, bed_key))

        log.print_progress(
            "Merge all of your annotation BED files into one BED file"
        )
        merge_bed_path = self.workspace / "merged_annotation.bed"
        merge_bed_files(
            merge_bed_path,
            bed_file_and_keys,
            self.num_proc,
            self.force_overwrite,
        )
        log.print_progress("Compress your BED file.")
        bed_gz_path = compress_using_bgzip(
            merge_bed_path, self.force_overwrite
        )

        log.print_progress("Make an index of your BED file.")
        bed_idx_path = index_using_tabix(bed_gz_path, self.force_overwrite)

        return bed_gz_path, bed_idx_path

    def _save_as_env(self, bed_gz_path: Path, bed_idx_path: Path):
        self.set_env("MERGED_BED", bed_gz_path)
        self.set_env("MERGED_BED_INDEX", bed_idx_path)
        self.save_env()



'''
CWAS DAWN analysis step

This step uses the DNM counts file and the output of the categorization step, 
and then conducts DAWN analysis for the categorized DNM.
'''

import argparse
from pathlib import Path
import os, glob
import pandas as pd
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.manifold import TSNE
from rpy2.robjects.packages import importr
from tqdm import tqdm
from scipy.stats import norm
import random
import polars as pl

from cwas.core.dawn.clustering import kmeans_cluster
from cwas.core.dawn.supernodeWGS import supernodeWGS_func, data_collection
from cwas.runnable import Runnable
from cwas.utils.check import check_is_file
from cwas.utils.check import check_is_dir
from cwas.utils.check import check_num_proc
from cwas.utils.log import print_arg, print_progress

class Dawn(Runnable):
    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self._eig_vector = None
        self._corr_mat = None
        self._permut_test = None
        self._category_set = None
        self._k_val = None
        self.kmeans_r = importr('stats').kmeans

    @staticmethod
    def _print_args(args: argparse.Namespace):
        print_arg(
            "No. worker processes for the DAWN",
            f"{args.num_proc: ,d}",
        )
        print_arg("Eigen vector file in input files", args.eig_vector_file)
        print_arg("Correlation matrix file in input files", args.corr_mat_file)
        print_arg("Permutation test file in input files", args.permut_test_file)
        print_arg("Category variant (or sample) counts file (burden test result) in input files: ", args.category_count_file)
        print_arg("Output directory", args.output_dir_path)
        print_arg("Thresholds of count / correlation / size: ", ", ".join(list(map(str, [args.count_threshold, args.corr_threshold, args.size_threshold]))))

    @staticmethod
    def _check_args_validity(args: argparse.Namespace):
        check_num_proc(args.num_proc)
        check_is_file(args.eig_vector_file)
        check_is_file(args.corr_mat_file)
        check_is_file(args.permut_test_file)
        check_is_file(args.category_count_file)
        check_is_dir(args.output_dir_path)

    
    @property
    def num_proc(self):
        return self.args.num_proc
    
    @property
    def input_dir_path(self) -> Path:
        return self.args.input_dir_path
    
    @property
    def eig_vector_file(self):
        return self.args.eig_vector_file
    
    @property
    def eig_vector(self):
        if self._eig_vector is None:
            self._eig_vector = pl.read_csv(self.eig_vector_file, separator='\t', has_header=False).to_pandas()
            self._eig_vector.index = self._eig_vector.iloc[:,0].tolist()
            self._eig_vector = self._eig_vector.iloc[:,1:]
        return self._eig_vector
    
    @property
    def corr_mat_file(self):
        return self.args.corr_mat_file

    @property
    def corr_mat(self):
        if self._corr_mat is None:
            self._corr_mat = pd.read_pickle(self.corr_mat_file)
            self._corr_mat = self._corr_mat.loc[self.category_set, self.category_set].astype(np.float64)
        return self._corr_mat

    @property
    def permut_test_file(self):
        return self.args.permut_test_file

    @property
    def permut_test(self):
        if self._permut_test is None:
            self._permut_test = pd.read_table(self.permut_test_file, compression='gzip', sep="\t")
        return self._permut_test
    
    @property
    def output_dir_path(self):
        return self.args.output_dir_path
    
    @property
    def k_range(self):
        return self.args.k_range
    
    @property
    def k_val(self):
        if self._k_val is None: # initialization
            if self.args.k_val is not None:
                self._k_val = self.args.k_val
            else: # self.args.k_val is None
                km_cluster = kmeans_cluster(self._tsne_out, self.seed)
                ## k 입력이 없으면 k_range 가지고 optimal k를 찾음, k_range는 default 값이 있으므로 user input이 없어도 적용됨
                print_progress("K is not seleted. Find the optimal K between the range ({})".format(self.k_range))
                output_name = os.path.join(self.output_dir_path, "{}_choose_K_silhouette_score_plot.pdf".format(self.tag)) # silhouette score plot (dawn output 1)
                self._k_val = km_cluster.optimal_k(self.k_range, output_name)

                print_progress("Find the optimal K = {}".format(self._k_val))
                
            ## k 입력이 있는 경우 k_range를 무시하고 k를 사용함, k와 k_range 값이 둘 다 있으면 k가 우선됨
            print_arg("K for K-means clustering algorithm", self._k_val)
        return self._k_val      

    @property
    def seed(self):
        return self.args.seed

    @property
    def tsne_method(self):
        return self.args.tsne_method

    @property
    def tag(self) -> str:
        return str(self.args.tag)

    @property
    def category_set(self):
        if self._category_set is None:
            self._category_set = self.eig_vector.index.tolist()
        return self._category_set

    @property
    def category_count(self):
        category_count_ = pd.read_table(self.args.category_count_file, sep="\t")
        return category_count_

    @property
    def count_threshold(self):
        return self.args.count_threshold

    @property
    def corr_threshold(self):
        return self.args.corr_threshold
    
    @property
    def size_threshold(self):
        return self.args.size_threshold

    def run(self):
        self.tsne_projection()
        self.kmeans_clustering()
        self.dawn_analysis()
        print_progress("Done")

    def tsne_projection(self): ## t-SNE and k-means clustering by optimal K
        random.seed(self.seed)
        print_progress("t-SNE projection for {}".format(self.eig_vector_file))

        # self.eig_vector = pd.read_table(self._eig_vector, compression='gzip')

        if type(self.eig_vector.iloc[0,0]) != np.float64:
            self.eig_vector = self.eig_vector.astype(np.float64)

        U_pick = self.eig_vector.iloc[:, 1:50]
        scaler = StandardScaler(with_mean=False, with_std=False)
        scale_factor = np.sqrt(np.sum(U_pick**2, axis=1))
        U_norm = scaler.fit_transform(U_pick.T / scale_factor).T

        tsne_res = TSNE(n_components=2, 
                        perplexity=30,
                        n_iter=500,
                        random_state=self.seed,
                        init='pca',
                        method=self.tsne_method,
                        n_jobs=self.num_proc).fit_transform(U_norm)

        self._U_norm = U_norm
        self._tsne_out = pd.DataFrame(tsne_res, columns=['t-SNE1', 't-SNE2'])

    def kmeans_clustering(self):
        random.seed(self.seed)
        km_cluster = kmeans_cluster(self._tsne_out, self.seed)

        print_progress("K-means clustering with {} clusters.".format(self.k_val))
        i_init_pt = km_cluster.center_init(self.k_val)

        # initial centers of clusters are given
        fit = self.kmeans_r(self._U_norm, centers=self._U_norm[i_init_pt,], iter_max=100)
        fit_res = dict(zip(fit.names, list(fit)))
        fit_res['annotation'] = self.category_set
        self._fit_res = fit_res
        
    def dawn_analysis(self):
        random.seed(self.seed)
        print_progress("DAWN analysis")
        clusters = list(self._fit_res['cluster'])

        supernodeWGS_process = supernodeWGS_func(self.corr_mat,
                                                 self._fit_res,
                                                 self.k_val,
                                                 self.num_proc,
                                                 self.output_dir_path,
                                                 self.tag,
                                                 self.seed,
                                                 verbose=True)
        
        ## create correlation matrix blocks
        max_val = int(self.k_val * (self.k_val-1)/2 + self.k_val)
        print_progress("[DAWN] Create correlation matrix blocks")
        for i in tqdm(range(1, max_val+1)):
            supernodeWGS_process.corr_mat_blocks(i)
        
        ## preparation for dawn
        # Load data, which contains the number of variants in each category.
        category_count_sub = self.category_count.loc[self.category_count['Category'].isin(self.category_set)]
        permut_test_sub = self.permut_test.loc[self.permut_test['Category'].isin(self.category_set)]

        print_progress("[DAWN] Preprocess data for the DAWN analysis")
        cat_count_permut, cluster_idx, cluster_size = supernodeWGS_process.dawn_preprocess(category_count_sub,
                                                                                           permut_test_sub,
                                                                                           self.count_threshold,
                                                                                           self.size_threshold)
        print_progress("[DAWN] Compute graph on correlation matrix and form graph")
        form_data = data_collection(path=os.path.join(self.output_dir_path, "supernodeWGS_results", "blocks_"+self.tag),
                                    cores=self.num_proc,
                                    max_cluster=self.k_val,
                                    verbose=True)
        cor_mat = form_data.form_correlation(k=cluster_idx)
        g = supernodeWGS_process.form_graph_from_correlation(cor_mat,
                                                             func=lambda x: x>self.corr_threshold,
                                                             k=cluster_idx)
        adj_mat = pd.DataFrame(np.array(g.get_adjacency().data), index=g.vs['name'], columns=g.vs['name'])
        adj_mat.to_csv(os.path.join(self.output_dir_path, "{}.ipvalue_fdr_igraph.csv".format(self.tag)), sep=",")
        
        
        print_progress("[DAWN] Compute the test statistics with z-scores")
        testvec_res = form_data.form_testvec(vec=cat_count_permut['P'],
                                             clustering=clusters,
                                             flag_vec=cat_count_permut['flag'],
                                             k=cluster_idx,
                                             sparse=True,
                                             sumabsv=5.25)
        zval_supernode = testvec_res['vec']
       
        print_progress("[DAWN] Compute the test statistics with relative risks")
        riskvec_res = form_data.form_testvec(vec=cat_count_permut['Relative_Risk'],
                                             clustering=clusters,
                                             flag_vec=cat_count_permut['flag2'],
                                             k=cluster_idx)
        tmp = riskvec_res['vec']
        tmp = tmp.copy()
        tmp.loc[np.isnan(tmp)] = 0
        risk_supernode = np.exp(tmp)
        
        
        ## apply dawn to clusters
        print_progress("[DAWN] Apply DAWN to clusters and calculate p-values and FDR")
        hmrf_res = supernodeWGS_process.hmrf(z=zval_supernode,
                                             adj=adj_mat,
                                             seedindex=[0 for x in cluster_idx],
                                             verbose=True)
        fdr = supernodeWGS_process.report_results(vec=cluster_idx,
                                                  posterior=1-hmrf_res['post'],
                                                  pvalue=1-norm.cdf(zval_supernode),
                                                  Iupdate=hmrf_res['Iupdate'])
        fdr.indicator = fdr.indicator.astype(int)
        #fdr.to_csv(os.path.join(self.output_dir_path, "{}.ipvalue_fdr.txt".format(self.tag)), sep="\t", index=False)
        
        cluster_pval = np.empty(len(zval_supernode)) * np.nan
        cluster_pval[np.where(risk_supernode >= 1)[0]] = 1 - norm.cdf(zval_supernode[np.where(risk_supernode >= 1)[0]])
        cluster_pval[np.where(risk_supernode < 1)[0]] = norm.cdf(zval_supernode[np.where(risk_supernode < 1)[0]])

        mat = pd.DataFrame({"Cluster.idx":g.vs['name'], "Pval.cluster":cluster_pval, "Risk":risk_supernode})
        mat.to_csv(os.path.join(self.output_dir_path, "{}.ipvalue_fdr_ipvalue_risk.csv".format(self.tag)), sep=",", index=False)

        ## save dawn plots and tables
        ### annotation table
        print_progress("[DAWN] Save DAWN clusters with annotations")
        tmp_idx = list(range(g.vcount()))
        cluster_indices = np.array(g.vs['name'])[tmp_idx].astype(int)
        assert all(np.array(sorted(cluster_indices)) == np.array(sorted(cluster_idx[tmp_idx])))

        annot = [[] for x in cluster_indices]

        for i in range(len(cluster_indices)):
            node_idx = testvec_res['index'][tmp_idx[i]]
            node_nam = np.array(self._fit_res['annotation'])[node_idx]

            assert len(set(np.array(clusters)[node_idx])) == 1, "Stop if they don't belong in the same cluster"
            assert len(node_idx) <= cluster_size[cluster_indices[i]-1], "Stop if the number of categories in the cluster is bigger than the size of the cluster."

            annot[i] = list(node_nam)

        max_len = max(list(map(len, annot)))

        # Create a matrix where the result will be stored.
        annot_mat = pd.DataFrame(index=range(max_len), columns=range(len(cluster_indices)))

        for i in range(len(cluster_indices)):
            annot_mat.iloc[:len(annot[i]), i] = annot[i]
        
        annot_mat.columns = cluster_indices
        annot_mat.to_csv(os.path.join(self.output_dir_path, "{}.cluster_annotation.csv".format(self.tag)), sep=",", index=False)
        
        ## dawn plots
        print_progress("[DAWN] Draw DAWN clusters network graph and save graph layout")
        supernodeWGS_process.dawn_plot(g, zval_supernode, annot)



import argparse
import multiprocessing as mp
import multiprocessing.shared_memory as shm
from functools import reduce
from itertools import product
from math import ceil
from pathlib import Path
import parmap
from cwas.core.common import chunk_list
from tqdm import tqdm
import re

import pandas as pd
import numpy as np
import yaml, pickle
from functools import partial

import cwas.utils.log as log
from cwas.core.categorization.categorizer import Categorizer
from cwas.core.categorization.parser import (
    parse_annotated_vcf,
    parse_gene_matrix,
)
from cwas.runnable import Runnable
from cwas.utils.check import check_num_proc
from cwas.utils.check import check_is_file
from cwas.utils.check import check_is_dir

lock = mp.Lock()

class Categorization(Runnable):
    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self._annotated_vcf = None
        self._gene_matrix = None
        self._category_domain = None
        self._annotated_vcf_groupby_sample = None
        self._sample_ids = None
        self._result = None
        self._correlation_matrix = None
        self._intersection_matrix = None
        self._shm = None

    @staticmethod
    def _print_args(args: argparse.Namespace):
        log.print_arg(
            "No. worker processes for the categorization",
            f"{args.num_proc: ,d}",
        )
        log.print_arg("Annotated VCF file", args.input_path)
        log.print_arg("Genereate a correlation matrix and an intersection matrix", (False if args.generate_matrix is None else True))

    @staticmethod
    def _check_args_validity(args: argparse.Namespace):
        check_num_proc(args.num_proc)
        check_is_file(args.input_path)
        check_is_dir(args.output_dir_path)

    @property
    def num_proc(self):
        return self.args.num_proc

    @property
    def input_path(self):
        return self.args.input_path.resolve()

    @property
    def output_dir_path(self):
        return self.args.output_dir_path.resolve()

    @property
    def generate_matrix(self):
        return self.args.generate_matrix

    @property
    def mis_info_key(self) -> str:
        return self.get_env("VEP_MIS_INFO_KEY")

    @property
    def mis_thres(self) -> float:
        return float(self.get_env("VEP_MIS_THRES"))

    @property
    def result_path(self) -> Path:
        f_name = re.sub(r'annotated\.vcf\.gz|annotated\.vcf', 'categorization_result.txt.gz', self.input_path.name)
        return Path(
            f"{self.output_dir_path}/" + 
            f"{f_name}"
        )

    @property
    def matrix_path(self) -> Path:
        f_name = re.sub(r'annotated\.vcf\.gz|annotated\.vcf', 'correlation_matrix.pkl', self.input_path.name)
        return Path(
            f"{self.output_dir_path}/" +
            f"{f_name}"
        )

    @property
    def intersection_matrix_path(self) -> Path:
        f_name = re.sub(r'annotated\.vcf\.gz|annotated\.vcf', 'intersection_matrix.pkl', self.input_path.name)
        return Path(
            f"{self.output_dir_path}/" +
            f"{f_name}"
        )

    @property
    def annotated_vcf(self) -> pd.DataFrame:
        if self._annotated_vcf is None:
            log.print_progress("Parse the annotated VCF")
            self._annotated_vcf = parse_annotated_vcf(
                Path(self.input_path)
            )
        return self._annotated_vcf

    @property
    def gene_matrix(self) -> dict:
        if self._gene_matrix is None:
            self._gene_matrix = parse_gene_matrix(
                Path(self.get_env("GENE_MATRIX"))
            )
        return self._gene_matrix

    @property
    def category_domain(self) -> dict:
        if self._category_domain is None:
            with Path(self.get_env("CATEGORY_DOMAIN")).open(
                "r"
            ) as category_domain_file:
                self._category_domain = yaml.safe_load(category_domain_file)
        return self._category_domain

    @property
    def categorizer(self) -> Categorizer:
        categorizer = Categorizer(self.category_domain, self.gene_matrix, self.mis_info_key, self.mis_thres)
        return categorizer

    @property
    def annotated_vcf_groupby_sample(self):
        if self._annotated_vcf_groupby_sample is None:
            self._annotated_vcf_groupby_sample = self.annotated_vcf.groupby(
                "SAMPLE"
            )
        return self._annotated_vcf_groupby_sample

    @property 
    def sample_ids(self) -> list:
        if self._sample_ids is None:
            self._sample_ids = list(self.annotated_vcf_groupby_sample.groups)
        return self._sample_ids

    @property
    def annotated_vcf_split_by_sample(self) -> list:
        return [
            self.annotated_vcf_groupby_sample.get_group(sample_id)
            for sample_id in self.sample_ids
        ]

    @property
    def redundant_categories(self) -> set:
        redundant_category_table = pd.read_table(
            self.get_env("REDUNDANT_CATEGORY")
        )
        return reduce(
            lambda x, y: x.union(y),
            [
                self._get_redundant_categories_from_row(row)
                for _, row in redundant_category_table.iterrows()
            ],
            set(),
        )

    def _get_redundant_categories_from_row(self, row: pd.Series) -> list:
        annotation_term_lists = {}
        for k, v in row.items():
            if v == "*":
                annotation_term_lists[k] = self._category_domain[k]
            else:
                annotation_term_lists[k] = [v]

        return {
            "_".join(combination)
            for combination in product(
                annotation_term_lists["variant_type"],
                annotation_term_lists["gene_set"],
                annotation_term_lists["functional_score"],
                annotation_term_lists["gencode"],
                annotation_term_lists["functional_annotation"],
            )
        }

    def run(self):
        self.categorize_vcf()
        self.remove_redundant_category()
        self.generate_correlation_matrix()
        self.save_result()
        self.update_env()
        log.print_progress("Done")

    def categorize_vcf(self):
        results_each_sample = (
            self.categorize_vcf_for_each_sample()
            if self.num_proc == 1
            else self.categorize_vcf_for_each_sample_with_mp()
        )
        log.print_progress("Organize the results")
        self._result = pd.DataFrame(results_each_sample).fillna(0)
        self._result = self._result.astype(int)
        self._result["SAMPLE"] = self.sample_ids
        self._result.set_index("SAMPLE", inplace=True)

    def remove_redundant_category(self):
        log.print_progress("Remove redundant categories from the result")
        self._result.drop(
            self.redundant_categories,
            axis="columns",
            inplace=True,
            errors="ignore",
        )
        log.print_progress(
            f"{len(self._result.columns):,d} categories have remained."
        )

    def categorize_vcf_for_each_sample(self):
        result = []
        for i, sample_vcf in enumerate(self.annotated_vcf_split_by_sample, 1):
            if i % 100 == 0:
                log.print_progress(f"Categorize variants of {i} samples")
            result.append(self.categorizer.categorize_variant(sample_vcf))
        return result

    def categorize_vcf_for_each_sample_with_mp(self):
        sample_vcfs = self.annotated_vcf_split_by_sample
        with mp.Pool(self.num_proc) as pool:
            log.print_progress("Categorize your input variants")
            return pool.map(
                self.categorizer.categorize_variant,
                sample_vcfs,
                chunksize=ceil(len(sample_vcfs) / self.num_proc),
            )

    def generate_correlation_matrix(self):
        if self.generate_matrix is None:
            return

        if self.generate_matrix == "sample":
            log.print_progress("Get an intersection matrix between categories using the number of samples")

            if self.num_proc == 1:
                intersection_matrix = self.process_columns_single(column_range = range(self._result.shape[1]), matrix=self._result)
            else:
                # Split the column range into evenly sized chunks based on the number of workers
                chunks = chunk_list(range(self._result.shape[1]), self.num_proc)
                result = parmap.map(self.process_columns, chunks, matrix=self._result, pm_pbar=True, pm_processes=self.num_proc)
                # Concatenate the count values
                intersection_matrix = pd.concat([pd.concat(chunk_results, axis=1) for chunk_results in result], axis=1)

        elif self.generate_matrix == "variant":
            log.print_progress("Get an intersection matrix between categories using the number of variants")
            intersection_matrix = (
                self.get_intersection_matrix(self.annotated_vcf, self.categorizer, self._result.columns)
                if self.num_proc == 1
                else self.get_intersection_matrix_with_mp()
            )
        
        diag_sqrt = np.sqrt(np.diag(intersection_matrix))
        log.print_progress("Calculate a correlation matrix")
        self._intersection_matrix = intersection_matrix
        self._correlation_matrix = intersection_matrix/np.outer(diag_sqrt, diag_sqrt)

    @staticmethod
    def process_columns(column_range, matrix: pd.DataFrame) -> list:
        results = []
    
        # Iterate over the column range
        for i in column_range:
            # Multiply the i-th column with values in the matrix
            df_multiplied = matrix.mul(matrix.iloc[:, i], axis=0)
            
            # Count the number of values greater than 0 in each column
            count_values_gt_zero = (df_multiplied > 0).sum(axis=0)
            
            # Assign the column name to count_values_gt_zero
            count_values_gt_zero.name = matrix.columns[i]
            
            results.append(count_values_gt_zero)
        
        return results

    @staticmethod
    def process_columns_single(column_range, matrix: pd.DataFrame) -> pd.DataFrame:
        # Initialize an empty DataFrame to store the concatenated results
        result = pd.DataFrame()

        # Define the progress bar
        pbar = tqdm(column_range, desc='Processing')

        # Perform the multiplication in a loop
        for i in pbar:
            # Multiply the i-th column with values in the matrix
            df_multiplied = matrix.mul(matrix.iloc[:, i], axis=0)
            
            # Count the number of values greater than 0 in each column
            count_values_gt_zero = (df_multiplied > 0).sum(axis=0)
            
            # Assign the column name to count_values_gt_zero
            count_values_gt_zero.name = matrix.columns[i]
            
            # Concatenate the count values to the 'result' DataFrame
            result = pd.concat([result, count_values_gt_zero], axis=1)

        # Close the progress bar
        pbar.close()
        
        return result

    def get_intersection_matrix_with_mp(self):
        ## use only one third of the cores to avoid memory error
        log.print_progress(f"This step will use only {self.num_proc//3 + 1} worker processes to avoid memory error")
        split_vcfs = np.array_split(self.annotated_vcf, self.num_proc//3 + 1)
        _get_intersection_matrix = partial(self.get_intersection_matrix,
                                           categorizer=self.categorizer, 
                                           categories=self._result.columns)
        
        with mp.Pool(self.num_proc//3 + 1) as pool:
            return sum(pool.map(
                _get_intersection_matrix,
                split_vcfs
            ))
        
    @staticmethod
    def get_intersection_matrix(annotated_vcf: pd.DataFrame, categorizer: Categorizer, categories: pd.Index): 
        return pd.DataFrame(
            categorizer.get_intersection(annotated_vcf), 
            index=categories, 
            columns=categories
        ).fillna(0).astype(int)

    def save_result(self):
        log.print_progress(f"Save the result to the file {self.result_path}")
        self._result.to_csv(self.result_path, sep="\t")
        if self._correlation_matrix is not None:
            log.print_progress("Save the intersection matrix to file")
            pickle.dump(self._intersection_matrix, open(self.intersection_matrix_path, 'wb'), protocol=5)
            log.print_progress("Save the correlation matrix to file")
            pickle.dump(self._correlation_matrix, open(self.matrix_path, 'wb'), protocol=5)

    def update_env(self):
        self.set_env("CATEGORIZATION_RESULT", self.result_path)
        self.save_env()



"""
Tests of the 'Configuration' step
"""
import random
from pathlib import Path

import pytest
from cwas.configuration import Configuration
from cwas.env import Env


@pytest.fixture(scope="module")
def vep_mock(cwas_workspace: Path):
    return cwas_workspace / "vep"


@pytest.fixture(scope="module", autouse=True)
def setup(cwas_workspace: Path, vep_mock: Path):
    cwas_workspace.mkdir()
    vep_mock.touch()
    vep_mock.chmod(0o755)
    set_env(cwas_workspace)


@pytest.fixture(scope="module", autouse=True)
def teardown(cwas_workspace: Path, vep_mock: Path):
    yield
    vep_mock.unlink()
    env = Env()
    env.reset()
    env.remove_file()
    for f in cwas_workspace.glob("*"):
        f.unlink()
    cwas_workspace.rmdir()


def set_env(cwas_workspace: Path):
    env = Env()
    env.set_env("CWAS_WORKSPACE", cwas_workspace)
    env.save()


@pytest.fixture
def cwas_config(
    annotation_dir: Path,
    annotation_key_conf: Path,
    gene_matrix: Path,
    vep_mock: Path,
    vep_cache: Path,
    vep_conserv: Path,
    vep_loftee: Path,
    vep_ances: Path,
    vep_gerp: Path,
    vep_msdb: Path,
    vep_mskey: str,
    vep_msthr: int,
):
    config = {
        "ANNOTATION_DATA_DIR": annotation_dir,
        "GENE_MATRIX": gene_matrix,
        "ANNOTATION_KEY_CONFIG": annotation_key_conf,
        "VEP": vep_mock,
        "VEP_CACHE_DIR": vep_cache,
        "VEP_CONSERVATION_FILE": vep_conserv,
        "VEP_LOFTEE": vep_loftee,
        "VEP_HUMAN_ANCESTOR_FA": vep_ances,
        "VEP_GERP_BIGWIG": vep_gerp,
        "VEP_MIS_DB": vep_msdb,
        "VEP_MIS_INFO_KEY": vep_mskey,
        "VEP_MIS_THRES": vep_msthr
    }
    return config


@pytest.fixture
def invalid_file_path(cwas_workspace: Path):
    filename = f"invalid-{random.randint(1, 1000000)}"
    return cwas_workspace / filename


@pytest.fixture
def create_cwas_config_file(cwas_workspace, cwas_config):
    _create_cwas_config_file(cwas_workspace, cwas_config)


@pytest.fixture
def create_incomplete_cwas_config_file(cwas_workspace, cwas_config):
    _unset_required_config(cwas_config)
    _create_cwas_config_file(cwas_workspace, cwas_config)


@pytest.fixture
def create_cwas_config_file_without_optional(cwas_workspace, cwas_config):
    _unset_optional_config(cwas_config)
    _create_cwas_config_file(cwas_workspace, cwas_config)


@pytest.fixture
def create_cwas_config_file_invalid_file_path(
    cwas_workspace, cwas_config, invalid_file_path
):
    _set_invalid_file_path(cwas_config, invalid_file_path)
    _create_cwas_config_file(cwas_workspace, cwas_config)


@pytest.fixture
def create_cwas_config_file_invalid_dir_path(
    cwas_workspace, cwas_config, invalid_file_path
):
    _set_invalid_dir_path(cwas_config, invalid_file_path)
    _create_cwas_config_file(cwas_workspace, cwas_config)


@pytest.fixture
def create_cwas_config_file_invalid_vep_path(
    cwas_workspace, cwas_config, invalid_file_path
):
    _set_invalid_vep_path(cwas_config, invalid_file_path)
    _create_cwas_config_file(cwas_workspace, cwas_config)


def _create_cwas_config_file(cwas_workspace, cwas_config):
    config_path = cwas_workspace / "configuration.txt"

    with config_path.open("w") as config_file:
        for k, v in cwas_config.items():
            print(f"{k}={str(v)}", file=config_file)


def _unset_required_config(cwas_config):
    random_config_key = random.choice(
        ["GENE_MATRIX", "ANNOTATION_DATA_DIR", "VEP"]
    )
    cwas_config[random_config_key] = ""


def _unset_optional_config(cwas_config):
    random_config_key = random.choice(
        ["ANNOTATION_KEY_CONFIG", "BIGWIG_CUTOFF_CONFIG"]
    )
    cwas_config[random_config_key] = ""


def _set_invalid_file_path(cwas_config, invalid_file_path):
    random_file_key = random.choice(
        ["GENE_MATRIX", "ANNOTATION_KEY_CONFIG", "BIGWIG_CUTOFF_CONFIG"]
    )
    cwas_config[random_file_key] = invalid_file_path


def _set_invalid_dir_path(cwas_config, invalid_dir_path):
    cwas_config["ANNOTATION_DATA_DIR"] = invalid_dir_path


def _set_invalid_vep_path(cwas_config, invalid_vep_path):
    cwas_config["VEP"] = invalid_vep_path


@pytest.fixture
def configuration_inst():
    inst = Configuration.get_instance()
    return inst


def test_run_configuration_with_incomplete(
    configuration_inst, create_incomplete_cwas_config_file,
):
    with pytest.raises(ValueError):
        configuration_inst.run()


def test_run_configuration(
    cwas_workspace, configuration_inst, create_cwas_config_file
):
    configuration_inst.run()
    _check_config_outputs(cwas_workspace)


def test_run_configuration_without_optional(
    cwas_workspace, configuration_inst, create_cwas_config_file_without_optional
):
    configuration_inst.run()
    _check_config_outputs(cwas_workspace)


def _check_config_outputs(cwas_workspace):
    data_dir_symlink = cwas_workspace / "annotation-data"
    gene_matrix_symlink = cwas_workspace / "gene_matrix.txt"
    bed_key_list = cwas_workspace / "annotation_keys.yaml"
    category_domain_list = cwas_workspace / "category_domain.yaml"
    redundant_category_table = cwas_workspace / "redundant_category.txt"

    assert data_dir_symlink.is_dir() and data_dir_symlink.is_symlink()
    assert gene_matrix_symlink.is_file() and gene_matrix_symlink.is_symlink()
    assert bed_key_list.is_file()
    assert category_domain_list.is_file()
    assert redundant_category_table.is_file()

    # Teardown
    data_dir_symlink.unlink()
    gene_matrix_symlink.unlink()
    bed_key_list.unlink()
    category_domain_list.unlink()
    redundant_category_table.unlink()


def test_run_configuration_with_incomplete(
    configuration_inst, create_incomplete_cwas_config_file,
):
    with pytest.raises(ValueError):
        configuration_inst.run()


def test_run_configuration_with_invalid_file_path(
    configuration_inst, create_cwas_config_file_invalid_file_path
):
    with pytest.raises(FileNotFoundError):
        configuration_inst.run()


def test_run_configuration_with_invalid_dir_path(
    configuration_inst, create_cwas_config_file_invalid_dir_path
):
    with pytest.raises(NotADirectoryError):
        configuration_inst.run()


def test_run_configuration_with_invalid_vep_path(
    configuration_inst, create_cwas_config_file_invalid_vep_path
):
    with pytest.raises(ValueError):
        configuration_inst.run()


def test_env_after_run_configuration(
    configuration_inst, create_cwas_config_file
):
    configuration_inst.run()

    env_keys = [
        "VEP",
        "CWAS_WORKSPACE",
        "ANNOTATION_DATA",
        "GENE_MATRIX",
        "ANNOTATION_BED_KEY",
        "ANNOTATION_BW_KEY",
        "ANNOTATION_BW_CUTOFF",
        "CATEGORY_DOMAIN",
        "REDUNDANT_CATEGORY",
    ]
    for env_key in env_keys:
        assert configuration_inst.get_env(env_key)


def test_get_inst_without_env():
    _make_env_empty()
    with pytest.raises(RuntimeError):
        Configuration.get_instance()


def _make_env_empty():
    env = Env()
    env.reset()
    env.save()



import argparse
import cwas.factory
import pytest

from cwas.runnable import Runnable


def test_make_class_name():
    assert cwas.factory.make_class_name("categorization") == "Categorization"
    assert cwas.factory.make_class_name("annotation") == "Annotation"
    assert cwas.factory.make_class_name("burden_test") == "BurdenTest"


def test_create_factory():
    factory_inst = cwas.factory.create("categorization")
    assert isinstance(factory_inst.argparser(), argparse.ArgumentParser)
    assert isinstance(factory_inst.runnable(None), Runnable)

    factory_inst = cwas.factory.create("binomial_test")
    assert isinstance(factory_inst.argparser(), argparse.ArgumentParser)
    assert isinstance(factory_inst.runnable(None), Runnable)


def test_create_factory_with_invalid_step():
    with pytest.raises(ValueError):
        cwas.factory.create("burden_test")



"""
Tests of the 'Start' step
"""
from pathlib import Path

import pytest
from cwas.start import Start


@pytest.fixture(scope="module")
def args(cwas_workspace: Path) -> list:
    return ["-w", str(cwas_workspace)]


@pytest.fixture(scope="module", autouse=True)
def setup(args: list):
    inst = Start.get_instance(args)
    inst.run()


@pytest.fixture(scope="module", autouse=True)
def teardown(cwas_workspace: Path, cwas_env_path: Path):
    yield
    cwas_env_path.unlink()
    for f in cwas_workspace.glob("*"):
        f.unlink()
    cwas_workspace.rmdir()


def test_initial_file_exist(cwas_env_path: Path, cwas_workspace: Path):
    cwas_config_path = cwas_workspace / "configuration.txt"
    assert cwas_env_path.is_file()
    assert cwas_workspace.is_dir()
    assert cwas_config_path.is_file()


def test_config_keys(cwas_workspace: Path):
    config_key_set = set()
    cwas_config_path = cwas_workspace / "configuration.txt"
    with cwas_config_path.open() as config_file:
        for line in config_file:
            config_key, _ = line.strip().split("=")
            config_key_set.add(config_key)

    expected_key_set = {
        "ANNOTATION_DATA_DIR",
        "GENE_MATRIX",
        "ANNOTATION_KEY_CONFIG",
        "BIGWIG_CUTOFF_CONFIG",
        "VEP",
    }
    assert config_key_set == expected_key_set


def test_init_without_args():
    inst = Start.get_instance()
    expect_default_workspace = Path.home() / ".cwas"
    actual_workspace = getattr(inst, "workspace")
    assert expect_default_workspace == actual_workspace



"""
Default category domain information for CWAS project.
"""
from copy import deepcopy

# This domain dictionary is incomplete.
# 'functional_score' domains and 'functional_annotation' domains are added by BED file keys.
# 'gene_set' domains are added by gene matrix.
_default_domains = {
    "variant_type": ["All", "SNV", "Indel"],
    "functional_score": ["All"],
    "gene_set": ["Any"],
    "gencode": [  # GENCODE annotation categories
        "Any",
        "CodingRegion",
        "FrameshiftRegion",
        "InFrameRegion",
        "SilentRegion",
        "PTVRegion",
        "DamagingMissenseRegion",
        "MissenseRegion",
        "NoncodingRegion",
        "SpliceSiteNoncanonRegion",
        "IntronRegion",
        "PromoterRegion",
        "IntergenicRegion",
        "UTRsRegion",
        "lincRnaRegion",
        "OtherTranscriptRegion",
    ],
    "functional_annotation": ["Any"],  # Custom annotation categories
}

_domain_types = list(_default_domains.keys())

# A category (domain combination) that includes a redundant domain pair will be
# excluded from CWAS analysis.
_redundant_domain_pairs = {
    ("variant_type", "gencode"): {
        ("All", "FrameshiftRegion"),
        ("All", "InFrameRegion"),
        ("All", "DamagingMissenseRegion"),
        ("All", "MissenseRegion"),
        ("All", "SilentRegion"),
        ("Indel", "DamagingMissenseRegion"),
        ("Indel", "MissenseRegion"),
        ("Indel", "SilentRegion"),
        ("SNV", "FrameshiftRegion"),
        ("SNV", "InFrameRegion"),
    },
    ("gene_set", "gencode"): {
        ("Any", "CodingRegion"),
        ("Any", "FrameshiftRegion"),
        ("Any", "InFrameRegion"),
        ("Any", "PTVRegion"),
        ("Any", "DamagingMissenseRegion"),
        ("Any", "MissenseRegion"),
        ("Any", "SilentRegion"),
        ("Any", "lincRnaRegion"),
        ("lincRNA", "Any")
    },
}


def get_default_domains() -> dict:
    return deepcopy(_default_domains)


def get_domain_types() -> list:
    return deepcopy(_domain_types)


def get_redundant_domain_pairs() -> dict:
    return deepcopy(_redundant_domain_pairs)



"""
This module includes functions for parsing files used in the CWAS
categorization step. By parsing those files, these functions make the
pandas.DataFrame objects that can be directly used in the categorization
algorithm.
"""
from io import TextIOWrapper
import pathlib
import re
import gzip

import numpy as np
import pandas as pd
from cwas.core.common import int_to_bit_arr
from cwas.utils.log import print_err


# TODO: Make the code much clearer
def parse_annotated_vcf(vcf_path: pathlib.Path) -> pd.DataFrame:
    """ Parse a Variant Calling File (VCF) that includes Variant Effect
    Predictor (VEP) and CWAS annotation information and make a
    pandas.DataFrame object listing annotated variants.
    """
    variant_col_names = []
    variant_rows = []  # Item: a list of values of each column
    csq_field_names = []  # CSQ is information from VEP
    annot_field_names = []  # Custom annotation field names

    # Read and parse the input VCF
    with gzip.open(vcf_path, "rb") as vep_vcf_file:
        for line_bytes in vep_vcf_file:
            line = line_bytes.decode("utf-8")
            if line.startswith("#"):  # Comments
                if line.startswith("##INFO=<ID=CSQ"):
                    csq_field_names = _parse_vcf_info_field(line)
                elif line.startswith("##INFO=<ID=ANNOT"):
                    annot_field_names = _parse_annot_field(line)
                elif line.startswith("#CHROM"):
                    variant_col_names = _parse_vcf_header_line(line)
            else:  # Rows of variant information follow the comments.
                assert variant_col_names, "The VCF does not have column names."
                assert csq_field_names, "The VCF does not have CSQ information."
                assert annot_field_names, (
                    "The VCF does not have annotation " "information."
                )
                variant_row = line.rstrip("\n").split("\t")
                variant_rows.append(variant_row)

    result = pd.DataFrame(variant_rows, columns=variant_col_names)
    try:
        info_df = _parse_info_column(
            result["INFO"], csq_field_names, annot_field_names
        )
    except KeyError:
        print_err(
            "The VCF does not have INFO column or "
            "the INFO values do not have expected field keys."
        )
        raise

    result.drop(columns="INFO", inplace=True)
    result = pd.concat([result, info_df], axis="columns")

    return result


def _parse_vcf_info_field(line):
    csq_line = line.rstrip('">\n')
    info_format_start_idx = re.search(r"Format: ", csq_line).span()[1]
    csq_field_names = csq_line[info_format_start_idx:].split("|")

    return csq_field_names


def _parse_annot_field(line):
    annot_line = line.rstrip('">\n')
    annot_field_str_idx = re.search(r"Key=", annot_line).span()[1]
    annot_field_names = annot_line[annot_field_str_idx:].split("|")

    return annot_field_names


def _parse_vcf_header_line(line):
    variant_col_names = line[1:].rstrip("\n").split("\t")
    return variant_col_names


def _parse_info_column(
    info_column: pd.Series, csq_field_names: list, annot_field_names: list
) -> pd.DataFrame:
    """ Parse the INFO column and make a pd.DataFrame object """
    info_values = info_column.values
    info_dicts = list(map(_parse_info_str, info_values))
    info_df = pd.DataFrame(info_dicts)
    csq_df = _parse_csq_column(info_df["CSQ"], csq_field_names)
    annot_df = _parse_annot_column(info_df["ANNOT"], annot_field_names)
    info_df.drop(columns=["CSQ", "ANNOT"], inplace=True)
    info_df = pd.concat([info_df, csq_df, annot_df], axis="columns")

    return info_df


def _parse_info_str(info_str: str) -> dict:
    """ Parse the string of the INFO field to make a dictionary """
    info_dict = {}
    key_value_pairs = info_str.split(";")

    for key_value_pair in key_value_pairs:
        key, value = key_value_pair.split("=", 1)
        info_dict[key] = value

    return info_dict


def _parse_csq_column(
    csq_column: pd.Series, csq_field_names: list
) -> pd.DataFrame:
    """ Parse the CSQ strings in the CSQ column and make a pd.DataFrame
    object
    """
    csq_values = csq_column.values
    csq_records = list(map(lambda csq_str: csq_str.split("|"), csq_values))
    csq_df = pd.DataFrame(csq_records, columns=csq_field_names)

    return csq_df


def _parse_annot_column(
    annot_column: pd.Series, annot_field_names: list
) -> pd.DataFrame:
    """ Parse the annotation integer in the ANNOT column and make a
    pd.DataFrame object
    """
    annot_ints = annot_column.values.astype(float)
    annot_field_cnt = len(annot_field_names)
    annot_records = list(
        map(
            lambda annot_int: int_to_bit_arr(annot_int, annot_field_cnt),
            annot_ints,
        )
    )
    annot_df = pd.DataFrame(annot_records, columns=annot_field_names)

    return annot_df


def parse_gene_matrix(gene_matrix_path: pathlib.Path) -> dict:
    """ Parse the gene matrix file and make a dictionary.
    The keys and values of the dictionary are gene symbols
    and a set of type names where the gene is associated,
    respectively.
    """
    with gene_matrix_path.open("r") as gene_matrix_file:
        return _parse_gene_matrix(gene_matrix_file)


def _parse_gene_matrix(gene_matrix_file: TextIOWrapper) -> dict:
    result = dict()
    header = gene_matrix_file.readline()
    all_gene_types = np.array(header.rstrip("\n").split("\t")[2:])

    for line in gene_matrix_file:
        _, gene_symbol, *gene_matrix_values = line.rstrip("\n").split("\t")
        gene_types = all_gene_types[np.array(gene_matrix_values) == "1"]
        result[gene_symbol] = set(gene_types)

    return result



class Category:
    def __init__(
        self, variant_type, gene_set, functional_score, gencode, functional_annotation
    ) -> None:
        self._variant_type = variant_type
        self._gene_set = gene_set
        self._functional_score = functional_score
        self._gencode = gencode
        self._functional_annotation = functional_annotation

    def __repr__(self):
        return (
            f"Category("
            f'"{self._variant_type}", '
            f'"{self._gene_set}", '
            f'"{self._functional_score}", '
            f'"{self._gencode}", '
            f'"{self._functional_annotation}")'
        )

    def __str__(self):
        """This representation is from An et al., 2018"""
        return "_".join(
            [
                self._variant_type,
                self._gene_set,
                self._functional_score,
                self._gencode,
                self._functional_annotation,
            ]
        )

    def __eq__(self, other):
        return (
            self._variant_type == other._variant_type
            and self._gene_set == other._gene_set
            and self._functional_score == other._functional_score
            and self._gencode == other._gencode
            and self._functional_annotation == other._functional_annotation
        )

    def __hash__(self):
        return hash(repr(self))

    def to_dict(self):
        return {
            "variant_type": self._variant_type,
            "gene_set": self._gene_set,
            "functional_score": self._functional_score,
            "gencode": self._gencode,
            "functional_annotation": self._functional_annotation,
        }

    @staticmethod
    def from_str(category_str):
        return Category(*category_str.split("_"))



from pathlib import Path

import pytest
from cwas.core.annotation.vep import VepCmdGenerator


@pytest.fixture(scope="module")
def vep_path(cwas_workspace) -> str:
    return str(cwas_workspace / "vep_mock")


@pytest.fixture(scope="module")
def input_vcf_path(cwas_workspace) -> str:
    return str(cwas_workspace / "test_vep.vcf")


@pytest.fixture(scope="module", autouse=True)
def setup(cwas_workspace, vep_path, input_vcf_path):
    cwas_workspace.mkdir()
    Path(vep_path).touch()
    Path(input_vcf_path).touch()


@pytest.fixture(scope="module", autouse=True)
def teardown(cwas_workspace, vep_path, input_vcf_path):
    yield
    Path(vep_path).unlink()
    Path(input_vcf_path).unlink()
    cwas_workspace.rmdir()


def test_init_vep(vep_path, input_vcf_path):
    vep_inst = VepCmdGenerator(vep_path, input_vcf_path)
    assert vep_inst.vep_path == vep_path
    assert vep_inst.input_vcf_path == input_vcf_path


def test_init_vep_with_invalid_vep_path(cwas_workspace, input_vcf_path):
    invalid_vep_path = str(cwas_workspace / "_vep_not_exists")
    with pytest.raises(FileNotFoundError):
        _ = VepCmdGenerator(invalid_vep_path, input_vcf_path)


def test_init_vep_with_no_arg():
    with pytest.raises(TypeError):
        _ = VepCmdGenerator()


def test_init_vep_with_no_vep(input_vcf_path):
    with pytest.raises(ValueError):
        _ = VepCmdGenerator(None, input_vcf_path)


def test_init_vep_with_no_input_vcf(vep_path):
    with pytest.raises(ValueError):
        _ = VepCmdGenerator(vep_path, None)


def test_init_vep_with_invalid_input_vcf(vep_path, cwas_workspace):
    invalid_vcf_path = str(cwas_workspace / "test_not_exists.vcf")
    with pytest.raises(FileNotFoundError):
        _ = VepCmdGenerator(vep_path, invalid_vcf_path)


def test_output_vcf_path(vep_path, input_vcf_path):
    vep_inst = VepCmdGenerator(vep_path, input_vcf_path)
    assert vep_inst.output_vcf_path == input_vcf_path.replace(
        ".vcf", ".vep.vcf"
    )


def test_cmd(vep_path, input_vcf_path):
    vep_inst = VepCmdGenerator(vep_path, input_vcf_path)
    assert vep_inst.cmd_str.startswith(vep_path)
    assert f"-i {input_vcf_path}" in vep_inst.cmd_str
    assert (
        f"-o {input_vcf_path.replace('.vcf', '.vep.vcf')}" in vep_inst.cmd_str
    )


def test_cmd_for_bw_custom_annotation(vep_path, input_vcf_path, annotation_dir):
    vep_inst = VepCmdGenerator(vep_path, input_vcf_path)
    bw_paths = [
        (bw_path, f"test{i + 1}")
        for i, bw_path in enumerate(annotation_dir.glob("*.bw"))
    ]
    for bw_path, bw_key in bw_paths:
        vep_inst.add_bw_custom_annotation(str(bw_path), bw_key)

    for bw_path, bw_key in bw_paths:
        assert (
            f"--custom {bw_path},{bw_key},bigwig,overlap,0" in vep_inst.cmd_str
        )



from io import StringIO

from cwas.core.categorization import parser
from pandas import Series


def test_parse_info_field():
    csq_field_names = ["Test1", "Test2", "Test3"]
    info_field = (
        f'##INFO=<ID=CSQ,Number=.,Type=String,Description="'
        f"Consequence annotations from Ensembl VEP. Format: "
        f"{'|'.join(csq_field_names)}\">"
    )
    assert parser._parse_vcf_info_field(info_field) == csq_field_names


def test_parse_annot_field():
    annot_field_names = ["ANNOT1", "ANNOT2", "ANNOT3"]
    annot_field = f"##INFO=<ID=ANNOT,Key={'|'.join(annot_field_names)}>"
    assert parser._parse_annot_field(annot_field) == annot_field_names


def test_parse_vcf_header_line():
    vcf_column_titles = [
        "CHROM",
        "POS",
        "ID",
        "REF",
        "ALT",
        "QUAL",
        "FILTER",
        "INFO",
    ]

    header_line = "\t".join(vcf_column_titles)
    assert parser._parse_vcf_header_line(f"#{header_line}") == vcf_column_titles


def test_parse_info_column():
    info_column = Series(
        [
            "SAMPLE=A;BATCH=1;CSQ=1|2|3;ANNOT=7",
            "SAMPLE=B;BATCH=1;CSQ=2|3|1;ANNOT=5",
            "SAMPLE=C;BATCH=1;CSQ=3|2|1;ANNOT=3",
            "SAMPLE=D;BATCH=1;CSQ=3|1|2;ANNOT=1",
        ]
    )
    csq_field_names = ["CSQ1", "CSQ2", "CSQ3"]
    annot_field_names = ["ANNOT1", "ANNOT2", "ANNOT3"]
    info_df = parser._parse_info_column(
        info_column, csq_field_names, annot_field_names
    )
    assert info_df["CSQ1"].to_list() == ["1", "2", "3", "3"]
    assert info_df["CSQ2"].to_list() == ["2", "3", "2", "1"]
    assert info_df["CSQ3"].to_list() == ["3", "1", "1", "2"]
    assert info_df["ANNOT1"].to_list() == [1, 1, 1, 1]
    assert info_df["ANNOT2"].to_list() == [1, 0, 1, 0]
    assert info_df["ANNOT3"].to_list() == [1, 1, 0, 0]


def test_parse_gene_matrix():
    gene_matrix_lines = [
        "\t".join(["gene_id", "gene_name", "asd1", "asd2", "asd3", "asd4"]),
        "\t".join(["id1", "gene1", "1", "0", "1", "1"]),
        "\t".join(["id2", "gene2", "0", "0", "1", "1"]),
        "\t".join(["id3", "gene3", "0", "1", "1", "0"]),
    ]
    result = parser._parse_gene_matrix(StringIO("\n".join(gene_matrix_lines)))
    assert result["gene1"] == {"asd1", "asd3", "asd4"}
    assert result["gene2"] == {"asd3", "asd4"}
    assert result["gene3"] == {"asd2", "asd3"}



"""
Test cwas.env
"""
from __future__ import annotations

import os
from pathlib import Path
from typing import OrderedDict

import pytest
from cwas.env import Env


@pytest.fixture(scope="module")
def env_inst():
    return Env()


@pytest.fixture(scope="module", autouse=True)
def teardown(env_inst: Env):
    yield
    env_inst.reset()
    env_inst.remove_file()


def test_env_init(env_inst: Env, cwas_env_path: Path):
    assert env_inst.get_path() is cwas_env_path
    assert isinstance(env_inst.env, OrderedDict)


def test_env_singleton(cwas_env_path: Path):
    env1 = Env()
    env2 = Env()
    assert env1 is env2

    new_env_path = Path(str(cwas_env_path) + "_new")
    env3 = Env(new_env_path)
    env4 = Env()
    assert env1 is env3
    assert env1 is env4
    assert env1.get_path() is new_env_path

    env1.set_path(cwas_env_path)


def test_env_setting(env_inst: Env):
    env_key = "TEST"
    expected = "Hello!"
    env_inst.set_env(env_key, expected)
    env_value = env_inst.get_env(env_key)
    assert env_value == expected


def test_env_get_nonexist(env_inst: Env):
    env_value = env_inst.get_env("FAKE")
    assert env_value is None


def test_env_prev_set_exists(env_inst: Env):
    # Test if the environment variable exists.
    env_key = "TEST"
    expected = "Hello!"
    env_value = env_inst.get_env(env_key)
    assert env_value == expected


def test_env_reset(env_inst: Env):
    env_inst.reset()
    env_key = "TEST"
    env_value = env_inst.get_env(env_key)
    assert not env_inst.env
    assert env_value is None


def test_env_save(env_inst: Env):
    env_key = "TEST"
    expected = "Hello!"
    env_inst.set_env(env_key, expected)
    env_inst.save()

    cwas_env_path = env_inst.get_path()
    with cwas_env_path.open() as env_file:
        env_line = env_file.read()
        env_line = env_line.strip()
    assert env_line == "TEST=Hello!"

    env_inst.get_path().unlink()


def test_load_env_to_os(env_inst: Env):
    env_key = "CWAS_TEST"
    env_value = "HELLO"
    env_inst.set_env(env_key, env_value)
    env_inst.save()

    env_inst.load_env_to_os()
    assert os.getenv(env_key) == env_value

    env_inst.get_path().unlink()
    os.environ.pop(env_key)


def test_load_env_to_os_without_save(env_inst: Env):
    env_key = "CWAS_TEST"
    env_value = "HELLO"
    env_inst.set_env(env_key, env_value)
    env_inst.load_env_to_os()
    assert os.getenv(env_key) is None


def test_load_env_from_file(env_inst: Env):
    env_inst.set_env("TEST", "TEST")
    env_inst.save()
    env_inst.reset()
    assert not env_inst.env
    env_inst.load_env_from_file()
    assert env_inst.env



import cwas.core.categorization.utils as utils


def test_get_idx_dict():
    assert utils.get_idx_dict(["a", "b", "c"]) == {"a": 0, "b": 1, "c": 2}


def test_extract_sublist_by_int():
    test_list = ["a", "b", "c", "d"]
    assert utils.extract_sublist_by_int(test_list, 0) == []
    assert utils.extract_sublist_by_int(test_list, 1) == ["a"]
    assert utils.extract_sublist_by_int(test_list, 2) == ["b"]
    assert utils.extract_sublist_by_int(test_list, 4) == ["c"]
    assert utils.extract_sublist_by_int(test_list, 8) == ["d"]
    assert utils.extract_sublist_by_int(test_list, 3) == ["a", "b"]
    assert utils.extract_sublist_by_int(test_list, 11) == ["a", "b", "d"]
    assert utils.extract_sublist_by_int(test_list, 12) == ["c", "d"]
    assert utils.extract_sublist_by_int(test_list, 13) == ["a", "c", "d"]
